from fastai.vision import Image
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
from fastai.torch_core import add_metrics
from torch.jit.annotations import List, Dict
from torchvision.models.detection.roi_heads import (
    fastrcnn_loss,
    maskrcnn_loss,
    maskrcnn_inference,
)
from torchvision.models.detection.transform import (
    resize_boxes,
    paste_masks_in_image,
)
import torch
from torch.nn.parallel import DistributedDataParallel
import warnings
import numpy as np
import matplotlib.pyplot as plt
from .._utils.common import ArcGISMSImage
from typing import Callable
import warnings
from .._utils.utils import check_imbalance
from fastprogress.fastprogress import progress_bar
from torchvision.ops import boxes as box_ops
from fastai.vision.transform import dihedral_affine
from fastai.vision import ImageBBox


def forward_roi(self, features, proposals, image_shapes, targets=None):
    """
    Arguments:
        features (List[Tensor])
        proposals (List[Tensor[N, 4]])
        image_shapes (List[Tuple[H, W]])
        targets (List[Dict])
    """

    train_val = getattr(self, "train_val", False)

    if targets is not None:
        for t in targets:
            floating_point_types = (torch.float, torch.double, torch.half)
            assert (
                t["boxes"].dtype in floating_point_types
            ), "target boxes must of float type"
            assert t["labels"].dtype == torch.int64, "target labels must of int64 type"

    losses = {}
    result = torch.jit.annotate(List[Dict[str, torch.Tensor]], [])

    if self.training:
        if train_val:
            original_prpsl = [p.clone() for p in proposals]
        (
            proposals,
            matched_idxs,
            labels,
            regression_targets,
        ) = self.select_training_samples(proposals, targets)

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)

        loss_classifier, loss_box_reg = fastrcnn_loss(
            class_logits, box_regression, labels, regression_targets
        )

        # during training, only focus on positive boxes
        num_images = len(proposals)
        mask_proposals = []
        pos_matched_idxs = []
        for img_id in range(num_images):
            pos = torch.where(labels[img_id] > 0)[0]
            mask_proposals.append(proposals[img_id][pos])
            pos_matched_idxs.append(matched_idxs[img_id][pos])

        mask_features = self.mask_roi_pool(features, mask_proposals, image_shapes)
        mask_features = self.mask_head(mask_features)
        mask_logits = self.mask_predictor(mask_features)

        gt_masks = [t["masks"] for t in targets]
        gt_labels = [t["labels"] for t in targets]
        rcnn_loss_mask = maskrcnn_loss(
            mask_logits, mask_proposals, gt_masks, gt_labels, pos_matched_idxs
        )

        losses = {
            "loss_classifier": loss_classifier,
            "loss_box_reg": loss_box_reg,
            "loss_mask": rcnn_loss_mask,
        }

    if not self.training or train_val:
        if train_val:
            proposals = original_prpsl

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)

        boxes, scores, labels = self.postprocess_detections(
            class_logits, box_regression, proposals, image_shapes
        )
        num_images = len(boxes)
        for i in range(num_images):
            result.append(
                {
                    "boxes": boxes[i],
                    "labels": labels[i],
                    "scores": scores[i],
                }
            )

        mask_proposals = [p["boxes"] for p in result]
        mask_features = self.mask_roi_pool(features, mask_proposals, image_shapes)
        mask_features = self.mask_head(mask_features)
        mask_logits = self.mask_predictor(mask_features)
        labels = [r["labels"] for r in result]
        masks_probs = maskrcnn_inference(mask_logits, labels)
        for mask_prob, r in zip(masks_probs, result):
            r["masks"] = mask_prob

    return result, losses


def postprocess_transform(self, result, image_shapes, original_image_sizes):
    train_val = getattr(self, "train_val", False)

    if not self.training or train_val:
        for i, (pred, im_s, o_im_s) in enumerate(
            zip(result, image_shapes, original_image_sizes)
        ):
            boxes = pred["boxes"]
            boxes = resize_boxes(boxes, im_s, o_im_s)
            result[i]["boxes"] = boxes
            if "masks" in pred:
                masks = pred["masks"]
                masks = paste_masks_in_image(masks, boxes, o_im_s)
                result[i]["masks"] = masks

    elif self.training:
        return result

    return result


def post_nms_top_n(self):
    train_val = getattr(self, "train_val", False)

    if train_val:
        return self._post_nms_top_n["testing"]
    elif self.training:
        return self._post_nms_top_n["training"]
    return self._post_nms_top_n["testing"]


def pre_nms_top_n(self):
    train_val = getattr(self, "train_val", False)

    if train_val:
        self._pre_nms_top_n["testing"]
    elif self.training:
        return self._pre_nms_top_n["training"]
    return self._pre_nms_top_n["testing"]


def eager_outputs_modified(self, losses, detections):
    train_val = getattr(self, "train_val", False)

    if train_val:
        return detections, losses
    elif self.training:
        return losses
    return detections


class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."

    def __init__(self, x, cmap=None, norm=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.cmap = cmap
        self.mplnorm = norm
        self.type = np.unique(x)

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs["mode"] = "nearest"
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(
        self,
        ax=None,
        figsize=(3, 3),
        title=None,
        hide_axis=True,
        cmap="tab20",
        alpha=0.5,
        **kwargs,
    ):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        masks = self.data[0].numpy()
        for i in range(1, self.data.shape[0]):
            max_unique = np.max(np.unique(masks))
            mask = np.where(self.data[i] > 0, self.data[i] + max_unique, self.data[i])
            masks += mask
        ax.imshow(masks, cmap=cmap, alpha=alpha, **kwargs)
        if hide_axis:
            ax.axis("off")
        if title:
            ax.set_title(title)


def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1.0, -1.0, -1.0]).any()


class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor

    def __init__(
        self,
        items,
        chip_size,
        classes=None,
        class_mapping=None,
        color_mapping=None,
        index_dir=None,
        **kwargs,
    ):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append("classes")
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        self.chip_size = chip_size
        self.inverse_class_mapping = {}
        self.index_dir = index_dir
        for k, v in self.class_mapping.items():
            self.inverse_class_mapping[v] = k
        if is_no_color(list(color_mapping.values())):
            self.cmap = "tab20"  ## compute cmap from palette
            import matplotlib as mpl

            bounds = list(color_mapping.keys())
            if (
                len(bounds) < 3
            ):  # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds) + 1]
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, len(bounds))
        else:
            import matplotlib as mpl

            bounds = list(color_mapping.keys())
            if (
                len(bounds) < 3
            ):  # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds) + 1]
            self.cmap = mpl.colors.ListedColormap(
                np.array(list(color_mapping.values())) / 255
            )
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, self.cmap.N)

        if len(color_mapping.keys()) == 1:
            self.cmap = "tab20"
            self.mplnorm = None

    def open(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)  # EXIF warning from TiffPlugin
            if len(fn) != 0:
                img_shape = ArcGISMSImage.read_image(fn[0]).shape
            else:
                labeled_mask = torch.zeros(
                    (len(self.class_mapping), self.chip_size, self.chip_size)
                )
                return ArcGISImageSegment(
                    labeled_mask, cmap=self.cmap, norm=self.mplnorm
                )
            k = 0

            labeled_mask = np.zeros((1, img_shape[0], img_shape[1]))

            for j in range(len(self.class_mapping)):
                if k < len(fn):
                    lbl_name = int(
                        self.index_dir[self.inverse_class_mapping[fn[k].parent.name]]
                    )
                else:
                    lbl_name = len(self.class_mapping) + 2
                if lbl_name == j + 1:
                    img = ArcGISMSImage.read_image(fn[k])
                    k = k + 1
                    if len(img.shape) == 3:
                        img = img.transpose(2, 0, 1)
                        img_mask = img[0]
                        for i in range(1, img.shape[0]):
                            max_unique = np.max(np.unique(img_mask))
                            img_i = np.where(img[i] > 0, img[i] + max_unique, img[i])
                            img_mask += img_i
                        img_mask = np.expand_dims(img_mask, axis=0)
                    else:
                        img_mask = np.expand_dims(img, axis=0)
                else:
                    img_mask = np.zeros((1, img_shape[0], img_shape[1]))
                labeled_mask = np.append(labeled_mask, img_mask, axis=0)
            labeled_mask = labeled_mask[1:, :, :]
            labeled_mask = torch.Tensor(list(labeled_mask))
        return ArcGISImageSegment(labeled_mask, cmap=self.cmap, norm=self.mplnorm)

    def reconstruct(self, t):
        return ArcGISImageSegment(t, cmap=self.cmap, norm=self.mplnorm)


class ArcGISInstanceSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False
    _div = None
    _imagery_type = None

    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

    def check_class_imbalance(
        self, func: Callable, stratify=False, class_imbalance_pct=0.01
    ):
        try:
            labelval = [(func(o)) for o in self.items]
            total_sample = np.concatenate(labelval)
            unique_sample = set(total_sample)

            check_imbalance(total_sample, unique_sample, class_imbalance_pct, stratify)
        except Exception as e:
            warnings.warn(f"Unable to check for class imbalance [reason : {e}]")

        return self

    def label_list_from_func(self, func: Callable):
        "Apply `func` to every input to get its label."
        import pandas as pd

        self._list_of_labels = ["_".join(func(o)) for o in self.items]
        self._idx_label_tuple_list = [
            (i, label) for i, label in enumerate(self._list_of_labels)
        ]
        label_series = pd.Series(self._list_of_labels)
        single_instance_labels = list(
            label_series.value_counts()[label_series.value_counts() == 1].index
        )
        self._label_idx_mapping = {
            label: i for i, label in enumerate(self._list_of_labels)
        }
        for (
            label
        ) in single_instance_labels:  # adding duplicate instance of unique labels
            self._idx_label_tuple_list.append((self._label_idx_mapping[label], label))
        return self

    def stratified_split_by_pct(self, valid_pct: float = 0.2, seed: int = None):
        try:
            "Split the items in a stratified manner by putting `valid_pct` in the validation set, optional `seed` can be passed."
            from sklearn.model_selection import train_test_split
            import random, math

            if valid_pct == 0.0:
                return self.split_none()
            if seed is not None:
                np.random.seed(seed)
            if (
                len(set(self._list_of_labels)) > len(self._list_of_labels) * valid_pct
            ):  # if validation samples length is less than unique labels
                classes = len(set(self._list_of_labels))
                xlen = len(self._list_of_labels)
                sample_shortage = math.ceil((classes - xlen * valid_pct) / valid_pct)
                extra_samples = random.choices(
                    self._idx_label_tuple_list, k=sample_shortage
                )
                self._idx_label_tuple_list.extend(extra_samples)
            X, y = [], []
            for index, label in self._idx_label_tuple_list:
                X.append(index)
                y.append(label)
            train_idx, val_idx, _, _ = train_test_split(
                X, y, test_size=valid_pct, random_state=seed, stratify=y
            )
            return self.split_by_idxs(train_idx, val_idx)
        except Exception as e:
            warnings.warn(
                f"Unable to perform stratified splitting [reason : {e}], falling back to random split"
            )
            return self.split_by_rand_pct(valid_pct=valid_pct, seed=seed)


class ArcGISInstanceSegmentationMSItemList(ArcGISInstanceSegmentationItemList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False

    def open(self, fn):
        return ArcGISMSImage.open_gdal(fn)


def mask_rcnn_loss(loss_value, *args):
    if isinstance(loss_value, tuple):
        loss_value = loss_value[1]

    final_loss = 0.0
    for i in loss_value.values():
        i[torch.isnan(i)] = 0.0
        i[torch.isinf(i)] = 0.0
        final_loss += i

    return final_loss


def mask_to_dict(last_target, device):
    target_list = []

    for i in range(len(last_target)):
        boxes = []
        masks = np.zeros((1, last_target[i].shape[1], last_target[i].shape[2]))
        labels = []
        for j in range(last_target[i].shape[0]):
            mask = np.array(last_target[i].data[j].cpu())
            obj_ids = np.unique(mask)

            if len(obj_ids) == 1:
                continue

            obj_ids = obj_ids[1:]
            mask_j = mask == obj_ids[:, None, None]
            num_objs = len(obj_ids)

            for k in range(num_objs):
                pos = np.where(mask_j[k])
                xmin = np.min(pos[1])
                xmax = np.max(pos[1])
                ymin = np.min(pos[0])
                ymax = np.max(pos[0])
                if xmax - xmin == 0:
                    xmax += 1
                if ymax - ymin == 0:
                    ymax += 1
                boxes.append([xmin, ymin, xmax, ymax])

            masks = np.append(masks, mask_j, axis=0)
            labels_j = torch.ones((num_objs,), dtype=torch.int64)
            labels_j = labels_j * (j + 1)
            labels.append(labels_j)

        if masks.shape[0] == 1:  # if no object in image
            masks[0, 50:51, 50:51] = 1
            labels = torch.tensor([0])
            boxes = torch.tensor([[50.0, 50.0, 51.0, 51.0]])
        else:
            labels = torch.cat(labels)
            boxes = torch.as_tensor(boxes, dtype=torch.float32)
            masks = masks[1:, :, :]
        masks = torch.as_tensor(masks, dtype=torch.uint8)
        target = {}
        target["boxes"] = boxes.to(device)
        target["labels"] = labels.to(device)
        target["masks"] = masks.to(device)
        target_list.append(target)

    return target_list


class train_callback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)

    def on_batch_begin(self, last_input, last_target, **kwargs):
        "Handle new batch `xb`,`yb` in `train` or validation."
        train = kwargs.get("train")
        if isinstance(self.model, DistributedDataParallel):
            model = self.model.module
        else:
            model = self.model
        self.model.train()
        if train:
            model.roi_heads.train_val = False
            model.rpn.train_val = False
            model.train_val = False
            model.transform.train_val = False
        else:
            model.backbone.eval()  # to get feature in eval mode for evaluation
            model.roi_heads.train_val = True
            model.rpn.train_val = True
            model.train_val = True
            model.transform.train_val = True
        target_list = mask_to_dict(last_target, self.c_device)
        if last_input.shape[0] < 2:
            last_input = torch.cat((last_input, last_input))
            target_list.append(target_list[0])

        last_input = [list(last_input), target_list]
        last_target = target_list  # [torch.tensor([1]) for i in last_target]
        return {"last_input": last_input, "last_target": last_target}


class AveragePrecision(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)

    def on_epoch_begin(self, **kwargs):
        self.aps = []

    def on_batch_end(self, last_output, last_target, **kwargs):
        last_output = last_output[0]
        for i in range(len(last_output)):
            last_output[i]["masks"] = last_output[i]["masks"].squeeze()
            if last_output[i]["masks"].shape[0] == 0:
                continue
            if len(last_output[i]["masks"].shape) == 2:
                last_output[i]["masks"] = last_output[i]["masks"][None]
            ap = compute_ap(
                last_target[i]["labels"],
                last_target[i]["masks"],
                last_output[i]["labels"],
                last_output[i]["scores"],
                last_output[i]["masks"],
            )
            self.aps.append(ap)

    def on_epoch_end(self, last_metrics, **kwargs):
        if isinstance(self.model, DistributedDataParallel):
            model = self.model.module
        else:
            model = self.model
        model.roi_heads.train_val = False
        model.rpn.train_val = False
        model.train_val = False
        model.transform.train_val = False
        if self.aps == []:
            self.aps.append(0.0)
        self.aps = torch.mean(torch.tensor(self.aps))
        return add_metrics(last_metrics, self.aps)


def masks_iou(masks1, masks2):
    # Mask R-CNN

    # The MIT License (MIT)

    # Copyright (c) 2017 Matterport, Inc.

    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:

    # The above copyright notice and this permission notice shall be included in
    # all copies or substantial portions of the Software.

    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    # THE SOFTWARE.

    # Method is based on https://github.com/matterport/Mask_RCNN

    if masks1.shape[0] == 0 or masks2.shape[0] == 0:
        return torch.zeros((masks1.shape[0], masks2.shape[0]))
    masks1 = masks1.permute(1, 2, 0)
    masks2 = masks2.permute(1, 2, 0)
    masks1 = torch.reshape(masks1 > 0.5, (-1, masks1.shape[-1])).type(torch.float64)
    masks2 = torch.reshape(masks2 > 0.5, (-1, masks2.shape[-1])).type(torch.float64)
    area1 = torch.sum(masks1, dim=0)
    area2 = torch.sum(masks2, dim=0)

    intersections = torch.mm(masks1.transpose(1, 0), masks2)
    union = area1[:, None] + area2[None, :] - intersections
    overlaps = intersections / union
    return overlaps


def compute_matches(
    gt_class_ids,
    gt_masks,
    pred_class_ids,
    pred_scores,
    pred_masks,
    iou_threshold=0.5,
    detect_threshold=0.5,
):
    # Method is based on https://github.com/matterport/Mask_RCNN
    indices = torch.argsort(pred_scores, descending=True)
    pred_class_ids = pred_class_ids[indices]
    pred_scores = pred_scores[indices]
    pred_masks = pred_masks[indices]

    ious_mask = masks_iou(pred_masks, gt_masks)

    pred_match = -1 * np.ones([pred_masks.shape[0]])
    if 0 not in ious_mask.shape:
        max_iou, matches = ious_mask.max(1)
        detected = []
        for i in range(len(pred_class_ids)):
            if (
                max_iou[i] >= iou_threshold
                and pred_scores[i] >= detect_threshold
                and matches[i] not in detected
                and gt_class_ids[matches[i]] == pred_class_ids[i]
            ):
                detected.append(matches[i])
                pred_match[i] = pred_class_ids[i]

    return pred_match


def compute_ap(
    gt_class_ids,
    gt_masks,
    pred_class_ids,
    pred_scores,
    pred_masks,
    iou_threshold=0.5,
    detect_threshold=0.5,
):
    # Method is based on https://github.com/matterport/Mask_RCNN
    pred_match = compute_matches(
        gt_class_ids,
        gt_masks,
        pred_class_ids,
        pred_scores,
        pred_masks,
        iou_threshold,
        detect_threshold,
    )

    precisions = np.cumsum(pred_match > -1) / (np.arange(len(pred_match)) + 1)
    recalls = np.cumsum(pred_match > -1).astype(np.float32) / len(gt_class_ids)

    precisions = np.concatenate([[0], precisions, [0]])
    recalls = np.concatenate([[0], recalls, [1]])

    for i in range(len(precisions) - 2, -1, -1):
        precisions[i] = np.maximum(precisions[i], precisions[i + 1])

    indices = np.where(recalls[:-1] != recalls[1:])[0] + 1
    mAP = np.sum((recalls[indices] - recalls[indices - 1]) * precisions[indices])
    return mAP


def batch_dihedral(x, k):
    flips = []
    if k & 1:
        flips.append(2)
    if k & 2:
        flips.append(3)
    if flips:
        x = torch.flip(x, flips)
    if k & 4:
        x = x.transpose(2, 3)
    return x.contiguous()


def recover_boxes(bboxes, size, k):
    if bboxes.size(0):
        device = bboxes.device
        bboxes = ImageBBox.create(*size, bboxes.detach().cpu())
        bboxes = dihedral_affine(bboxes, k)
        if k == 5 or k == 6:
            bboxes = dihedral_affine(bboxes, 3)
        bboxes = (size[0] * (bboxes.data + 1) / 2).to(device)

    return bboxes


def intersect(box_a, box_b):
    max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
    min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
    inter = torch.clamp((max_xy - min_xy), min=0)
    return inter[:, :, 0] * inter[:, :, 1]


def box_area(b):
    return (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])


def boxious(box_a, box_b):
    inter = intersect(box_a, box_b)
    union = box_area(box_a).unsqueeze(1) + box_area(box_b).unsqueeze(0) - inter
    return inter / union


def pred_mean_merge(pred, iou_thresold=0.5, same_pred=1):
    bboxes = pred["boxes"]
    masks = pred["masks"].squeeze()
    scores, labels = pred["scores"], pred["labels"]
    ious = boxious(bboxes, bboxes)
    numOfboxes = ious.shape[0]
    # merge predictions with iou>thresold
    ref_boxs, merge_boxids = np.where(ious > iou_thresold)
    _, idx = np.unique(ref_boxs, return_index=True)
    counter = np.array(range(numOfboxes))
    i = 0
    merge_masks, merge_boxes, merge_scores, merge_labels = [], [], [], []
    while i < numOfboxes:
        if counter[i] == -1:
            i += 1
            continue
        if i == numOfboxes - 1:
            matched_idx = merge_boxids[idx[i] :]
        else:
            matched_idx = merge_boxids[idx[i] : idx[i + 1]]
        if len(matched_idx) > same_pred:
            merge_masks.append(masks[matched_idx, :, :].mean(dim=0))
            merge_boxes.append(bboxes[matched_idx, :].mean(dim=0))
            merge_scores.append(scores[matched_idx].mean())
            merge_labels.append(labels[matched_idx[0]])

        counter[np.isin(counter, matched_idx)] = -1
        i += 1
    return (
        torch.stack(merge_masks) if merge_masks else torch.tensor([]),
        torch.stack(merge_boxes) if merge_boxes else torch.tensor([]),
        torch.tensor(merge_scores),
        torch.tensor(merge_labels),
    )


def merge_tta_prediction(predictions, nms_thres=0.3, merge_policy="mean"):
    device = predictions[0]["boxes"].device
    result = {k: [] for k in predictions[0].keys()}
    for pred in predictions:
        for k, v in pred.items():
            result[k].append(v)

    for k, v in result.items():
        result[k] = torch.cat(v).detach().cpu()

    if merge_policy == "mean":
        (
            result["masks"],
            result["boxes"],
            result["scores"],
            result["labels"],
        ) = pred_mean_merge(result, nms_thres)
    else:
        keep = box_ops.batched_nms(
            result["boxes"],
            result["scores"],
            result["labels"],
            nms_thres,
        )

        for k, v in result.items():
            result[k] = result[k][keep]

    for k, v in result.items():
        result[k] = result[k].to(device)

    return result


def predict_tta(model, batch, detect_thresh=0.5, merge_policy="mean"):
    temp = model.roi_heads.score_thresh
    model.roi_heads.score_thresh = detect_thresh
    ttaPreds = [[] for _ in range(batch.shape[0])]
    for k in model.arcgis_tta:
        transforms = batch_dihedral(batch, k)
        pred = model(list(transforms))
        for i, p in enumerate(pred):
            p["masks"] = batch_dihedral(p["masks"], k)
            if k == 5 or k == 6:
                p["masks"] = batch_dihedral(p["masks"], 3)
            p["boxes"] = recover_boxes(p["boxes"], batch.shape[-2:], k)
            ttaPreds[i].append(p)
    for i, pred in enumerate(ttaPreds):
        ttaPreds[i] = merge_tta_prediction(
            pred, model.roi_heads.nms_thresh, merge_policy
        )
    model.roi_heads.score_thresh = temp

    return ttaPreds


def compute_class_AP(
    model,
    dl,
    n_classes,
    show_progress,
    detect_thresh=0.5,
    iou_thresh=0.5,
    mean=False,
    tta_prediction=False,
):
    model.learn.model.eval()
    if mean:
        aps = []
    else:
        aps = [[] for _ in range(n_classes)]
    with torch.no_grad():
        for input, target in progress_bar(dl, display=show_progress):
            if tta_prediction:
                predictions = predict_tta(model.learn.model, input, detect_thresh)
            else:
                predictions = model.learn.model(list(input))
            ground_truth = mask_to_dict(target, model._device)
            for i in range(len(predictions)):
                predictions[i]["masks"] = predictions[i]["masks"].squeeze()
                if predictions[i]["masks"].shape[0] == 0:
                    continue
                if len(predictions[i]["masks"].shape) == 2:
                    predictions[i]["masks"] = predictions[i]["masks"][None]
                if mean:
                    ap = compute_ap(
                        ground_truth[i]["labels"],
                        ground_truth[i]["masks"],
                        predictions[i]["labels"],
                        predictions[i]["scores"],
                        predictions[i]["masks"],
                        iou_thresh,
                        detect_thresh,
                    )
                    aps.append(ap)
                else:
                    for k in range(1, n_classes + 1):
                        gt_labels_index = (
                            (ground_truth[i]["labels"] == k).nonzero().reshape(-1)
                        )
                        gt_labels = ground_truth[i]["labels"][gt_labels_index]
                        gt_masks = ground_truth[i]["masks"][gt_labels_index]
                        pred_labels_index = (
                            (predictions[i]["labels"] == k).nonzero().reshape(-1)
                        )
                        pred_labels = predictions[i]["labels"][pred_labels_index]
                        pred_masks = predictions[i]["masks"][pred_labels_index]
                        pred_scores = predictions[i]["scores"][pred_labels_index]
                        if len(gt_labels):
                            ap = compute_ap(
                                gt_labels,
                                gt_masks,
                                pred_labels,
                                pred_scores,
                                pred_masks,
                                iou_thresh,
                                detect_thresh,
                            )
                            aps[k - 1].append(ap)
    if mean:
        if aps != []:
            aps = np.mean(aps, axis=0)
        else:
            return 0.0
    else:
        for i in range(n_classes):
            if aps[i] != []:
                aps[i] = np.mean(aps[i])
            else:
                aps[i] = 0.0
    if model._device == torch.device("cuda"):
        torch.cuda.empty_cache()
    return aps



#!/usr/bin/env python3

"""
The methods are taken from => https://github.com/anilbatra2185/road_connectivity

MIT License
==========================================================================

Copyright (c) 2020 Anil Batra

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import math

import numpy as np

from .rdp import rdp


def simplify_edge(ps, max_distance=1):
    """
    Combine multiple points of graph edges to line segments
    so distance from points to segments <= max_distance

    @param ps: array of points in the edge, including node coordinates
    @param max_distance: maximum distance, if exceeded new segment started

    @return: ndarray of new nodes coordinates
    """
    res_points = []
    cur_idx = 0
    for i in range(1, len(ps) - 1):
        segment = ps[cur_idx : i + 1, :] - ps[cur_idx, :]
        angle = -math.atan2(segment[-1, 1], segment[-1, 0])
        ca = math.cos(angle)
        sa = math.sin(angle)
        # rotate all the points so line is alongside first column coordinate
        # and the second col coordinate means the distance to the line
        segment_rotated = np.array([[ca, -sa], [sa, ca]]).dot(segment.T)
        distance = np.max(np.abs(segment_rotated[1, :]))
        if distance > max_distance:
            res_points.append(ps[cur_idx, :])
            cur_idx = i
    if len(res_points) == 0:
        res_points.append(ps[0, :])
    res_points.append(ps[-1, :])

    return np.array(res_points)


def simplify_graph(graph, max_distance=1):
    """
    @params graph: MultiGraph object of networkx
    @return: simplified graph after applying RDP algorithm.
    """
    all_segments = []
    # Iterate over Graph Edges
    for s, e in graph.edges():
        for _, val in graph[s][e].items():
            # get all pixel points i.e. (x,y) between the edge
            ps = val["pts"]
            # create a full segment
            full_segments = np.row_stack([graph.nodes[s]["o"], ps, graph.nodes[e]["o"]])
            # simply the graph.
            segments = rdp(full_segments.tolist(), max_distance)
            all_segments.append(segments)

    return all_segments


def segment_to_linestring(segment):
    """
    Convert Graph segment to LineString require to calculate the APLS mteric
    using utility tool provided by Spacenet.
    """

    if len(segment) < 2:
        return []
    linestring = "LINESTRING ({})"
    sublinestring = ""
    for i, node in enumerate(segment):
        if i == 0:
            sublinestring = sublinestring + "{:.1f} {:.1f}".format(node[1], node[0])
        else:
            if node[0] == segment[i - 1][0] and node[1] == segment[i - 1][1]:
                if len(segment) == 2:
                    return []
                continue
            if i > 1 and node[0] == segment[i - 2][0] and node[1] == segment[i - 2][1]:
                continue
            sublinestring = sublinestring + ", {:.1f} {:.1f}".format(node[1], node[0])
    linestring = linestring.format(sublinestring)
    return linestring


def segmets_to_linestrings(segments):
    """
    Convert multiple segments to LineStrings require to calculate the APLS mteric
    using utility tool provided by Spacenet.
    """

    linestrings = []
    for segment in segments:
        linestring = segment_to_linestring(segment)
        if len(linestring) > 0:
            linestrings.append(linestring)
    if len(linestrings) == 0:
        linestrings = ["LINESTRING EMPTY"]
    return linestrings


def unique(list1):
    # intilize a null list
    unique_list = []

    # traverse for all elements
    for x in list1:
        # check if exists in unique_list or not
        if x not in unique_list:
            unique_list.append(x)
    return unique_list



from __future__ import annotations
import hmac
import time
import base64
import struct
import typing

import urllib.parse as urllib_parse
import urllib.request
from functools import lru_cache
import importlib


def hotp(key: str, counter: int, digits: int = 6, digest: str = "sha1"):
    key = base64.b32decode(key.upper() + "=" * ((8 - len(key)) % 8))
    counter = struct.pack(">Q", counter)
    mac = hmac.new(key, counter, digest).digest()
    offset = mac[-1] & 0x0F
    binary = struct.unpack(">L", mac[offset : offset + 4])[0] & 0x7FFFFFFF
    return str(binary)[-digits:].zfill(digits)


def mfa_otp(
    key: str, time_step: int = 30, digits: int = 6, digest: str = "sha1"
) -> str:
    """Creates the MFA Code for MFA logins"""
    return hotp(key, int(time.time() / time_step), digits, digest)


@lru_cache(maxsize=255)
def check_module_exists(name: str) -> bool:
    """Checks if a module exists"""
    try:
        res = importlib.util.find_spec(name)
        if res is None:
            return False
        return True
    except:
        return False


@lru_cache(maxsize=255)
def parse_url(url: str) -> object:
    """
    Parses a URL string into it's pieces.

    :returns: Named Tuple
    """
    return urllib_parse.urlparse(url)


@lru_cache(maxsize=255)
def assemble_url(parsed: object) -> str:
    """
    creates the URL from a parsed URL
    """
    if parsed.port:
        netloc: str = parsed.netloc.split(":")[0]
        server_url = (
            f'{parsed.scheme}://{netloc}:{parsed.port}/{parsed.path[1:].split("/")[0]}'
        )
    else:
        server_url = (
            f'{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split("/")[0]}'
        )
    return server_url


def detect_proxy(replace_https: bool = True) -> typing.Optional[typing.Dict]:
    """
    Using `urllib.request.getproxies` create the dictionary for the proxy if they exist.

    .. note::
        This method cannot detect all proxies and is only recommended to assist if proxy
        errors occur. Talk to whomever manages your proxy to get the proper forward
        proxy information.

    .. note::
        See the urllib.requests.getproxies() `documentation
        <https://docs.python.org/3/library/urllib.request.html#urllib.request.getproxies>`_
         page for a full explination of the code.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    replace_https       Optional Boolean.  The autodetect method from `urllib.requests.getproxies`
                        assumes there is an `http` and `https` version of the proxy.  Many
                        implementations of proxies force `https` traffic through the `http`
                        proxy endpoint.  When this is true, the `getproxies` call will be
                        modified for the `https` entry and switch the scheme from `https`
                        to `http`.
    ===============     ====================================================================

    Usage Example:

       gis = GIS(proxy=detect_proxy(True), verify_cert=False)
       gis = GIS(proxy=detect_proxy(False), verify_cert=False)

    :returns: Dict[str,str] or None
    """
    proxies = urllib.request.getproxies()
    if proxies and replace_https:
        if "https" in proxies:
            proxies["https"] = proxies["https"].replace("https", "http")
        return proxies
    elif proxies and replace_https == False:
        return proxies
    elif proxies == {}:
        proxies = None
    else:
        proxies = None
    return proxies


def merge_proxies(
    proxy_dict: typing.Dict[str, str] = None,
    proxy_host: str = None,
    proxy_port: str = "8888",
    detect: bool = False,
    replace_https: bool = False,
) -> typing.Dict[str, str]:
    """
    The `merge_proxies` combines multiple proxies based on the Python
    API's logical workflow.  The `proxy_host` and `proxy_port` are first
    considered, then the system detected proxies from
    `urllib.request.getproxies`, and finally `proxy_dict`.  These
    dictionaries and then merged together.  If the save `keys` exist, then
    they are overwritten.


    .. note::

        Helper utility to assist with debugging/trouble shooting proxy issues.
        This method is not intended to be used for public consumption and may
        change in the future.

    .. code-block:: python

        # Usage Example: proxy_dict, host/port and do not detect proxies

        >>> proxy_dict = {'https' : 'https://127.0.0.1:8734'}
        >>> proxy_host = "localhost"
        >>> proxy_host = "9000"

        >>> print(merge_proxies(proxy_dict=proxy_dict, proxy_host=proxy_host, proxy_port=proxy_port))
        {'https' : 'https://127.0.0.1:8734', 'http' : "http://localhost:9000"}


    :returns: Dict[str,str]

    """
    if proxy_dict is None and proxy_host is None and detect == False:
        return None
    if proxy_host and proxy_port:
        default_proxies = {
            "http": f"http://{proxy_host}:{proxy_port}",
            "https": f"https://{proxy_host}:{proxy_port}",
        }
    else:
        default_proxies = {}
    if detect:
        dectected_proxies = detect_proxy(replace_https)
    else:
        dectected_proxies = {}
    if proxy_dict is None:
        proxy_dict = {}

    default_proxies.update(dectected_proxies)
    default_proxies.update(proxy_dict)
    if default_proxies == {}:
        return None
    return default_proxies



""" Defines the Integration class.
"""

from .exceptions import ValidationError
from .feature_model import FeatureModel
from ._store import *
from ._schemas import IntegrationSchema
from .assignment_type import AssignmentType


class Integration(FeatureModel):
    """
    Represents an integration in a project. Version 2 Workforce projects only.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                           the dispatcher belongs to.
    ------------------     --------------------------------------------------------------------
    feature                Optional :class:`~arcgis.features.Feature`. The feature representing
                           the dispatcher. Mostly intended for
                           internal usage. If supplied, other parameters are ignored.
    ------------------     --------------------------------------------------------------------
    integration_id         Optional :class:`String`. The id for the integration
    ------------------     --------------------------------------------------------------------
    prompt                 Optional :class:`String`. The prompt in the mobile app for the
                           integration
    ------------------     --------------------------------------------------------------------
    url_template           Optional :class:`String`. The url that the prompt links to
    ------------------     --------------------------------------------------------------------
    assignment_type        Optional :class:`String`. The assignment type for the integration
    ==================     ====================================================================

    .. code-block:: python

        # Get an integration, update it, delete it

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        integration = project.integrations.search()[0]
        integration.update(integration_id="arcgis-navigator",prompt="Navigate to Assignment")
        integration.delete()


    """

    def __init__(
        self,
        project,
        feature=None,
        integration_id=None,
        prompt=None,
        url_template=None,
        assignment_type=None,
    ):
        super().__init__(project, project.integrations_table, feature)
        self._schema = IntegrationSchema(project.integrations_table)
        if not feature:
            self.integration_id = integration_id
            self.prompt = prompt
            self.url_template = url_template
            if assignment_type:
                if isinstance(assignment_type, AssignmentType):
                    self.assignment_type = assignment_type.code
                else:
                    self.assignment_type = assignment_type.upper()

    def __str__(self):
        return "<Integration {}>".format(self.integration_id)

    def __repr__(self):
        return "<Integration {}>".format(self.integration_id)

    def update(
        self, integration_id=None, prompt=None, url_template=None, assignment_type=None
    ):
        """
        Updates the dispatcher on the server

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        integration_id         Optional :class:`String`. The id for the integration
        ------------------     --------------------------------------------------------------------
        prompt                 Optional :class:`String`. The prompt in the mobile app for the
                               integration
        ------------------     --------------------------------------------------------------------
        url_template           Optional :class:`String`. The url that the prompt links to
        ------------------     --------------------------------------------------------------------
        assignment_type        Optional :class:`String`. The assignment type for the integration
        ==================     ====================================================================

        """
        update_integration(
            self.project, self, integration_id, prompt, url_template, assignment_type
        )

    def delete(self):
        """Deletes the integration from the server"""
        delete_integrations(self.project, [self])

    @property
    def integration_id(self):
        """Gets/Sets the id of the integration"""
        return self._feature.attributes.get(self._schema.integration_id)

    @integration_id.setter
    def integration_id(self, value):
        self._feature.attributes[self._schema.integration_id] = value

    @property
    def prompt(self):
        """Gets/Sets the prompt of the integration"""
        return self._feature.attributes.get(self._schema.prompt)

    @prompt.setter
    def prompt(self, value):
        self._feature.attributes[self._schema.prompt] = value

    @property
    def url_template(self):
        """Gets/Sets the url template of the integration"""
        return self._feature.attributes.get(self._schema.url_template)

    @url_template.setter
    def url_template(self, value):
        self._feature.attributes[self._schema.url_template] = value

    @property
    def assignment_type(self):
        """Gets/Sets the assignment type of the integration"""
        return self._feature.attributes.get(self._schema.assignment_type)

    @assignment_type.setter
    def assignment_type(self, value):
        self._feature.attributes[self._schema.assignment_type] = value

    def _validate(self):
        if not self.integration_id:
            raise ValidationError("Integration must have an id", self)
        elif not self.prompt:
            raise ValidationError("Assignment integration must contain a prompt", self)
        elif not self.url_template:
            raise ValidationError(
                "Assignment integration must contain a URL template", self
            )
        elif self.assignment_type:
            if self.assignment_type.upper() not in [
                at.code.upper() for at in self.project.assignment_types.search()
            ]:
                raise ValidationError("Invalid assignment type in integration", self)
            # Don't enforce this when doing an update (which means the integration already has a global_id)
            for integration in self.project.integrations.search():
                if (
                    self.global_id is None
                    and integration.integration_id == self.integration_id
                    and not integration.assignment_type
                ):
                    raise ValidationError(
                        "Cannot add an integration with an assignment type when project level integration of same id exists",
                        self,
                    )
                if (
                    self.global_id is None
                    and integration.integration_id == self.integration_id
                    and integration.assignment_type.upper()
                    == self.assignment_type.upper()
                ):
                    raise ValidationError(
                        "Cannot add an integration with the same id and assignment type",
                        self,
                    )
        else:
            # Don't enforce this when doing an update (which means the integration already has a global_id)
            if self.global_id is None and self.integration_id in [
                integration.integration_id
                for integration in self.project.integrations.search()
            ]:
                raise ValidationError(
                    "Cannot add project level integration when same id integration exists",
                    self,
                )



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# retinanet_ghm_x101_64x4d_fpn_1x_coco.py, box AP=41.4
_base_ = "./_base_/models/retinanet_r50_fpn.py"
model = dict(
    pretrained="open-mmlab://resnext101_64x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=64,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        style="pytorch",
    ),
    bbox_head=dict(
        loss_cls=dict(
            _delete_=True,
            type="GHMC",
            bins=30,
            momentum=0.75,
            use_sigmoid=True,
            loss_weight=1.0,
        ),
        loss_bbox=dict(
            _delete_=True, type="GHMR", mu=0.02, bins=10, momentum=0.7, loss_weight=10.0
        ),
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/ghm/retinanet_ghm_x101_64x4d_fpn_1x_coco/retinanet_ghm_x101_64x4d_fpn_1x_coco_20200131-dd381cef.pth"



# Code from https://github.com/fastai/course-v3/blob/master/nbs/dl2/cyclegan_ws.ipynb & https://github.com/eriklindernoren/PyTorch-GAN
import torch
from fastai.vision import (
    nn,
    Callable,
    List,
    LearnerCallback,
    optim,
    ifnone,
    F,
    flatten_model,
    requires_grad,
    SmoothenValue,
    add_metrics,
)
from .._utils.cyclegan import (
    calculate_activation_statistics,
    calculate_frechet_distance,
)
from fastprogress.fastprogress import progress_bar
from .._utils.superres import psnr, ssim

import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from fastai.vision.learner import create_body
from torchvision.models.resnet import (
    resnet18,
    resnet34,
    resnet50,
    resnext50_32x4d,
    wide_resnet50_2,
)
from fastai.vision.models.unet import DynamicUnet


def GeneratorResUnet(data, n_output, size, backbone):
    from ._arcgis_model import _change_tail

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    body = create_body(backbone, pretrained=True)
    new_body = _change_tail(body, data)
    net_G = DynamicUnet(new_body, n_output, (size, size)).to(device)
    return net_G


##############################
#           U-NET
##############################


class UNetDown(nn.Module):
    def __init__(self, in_size, out_size, normalize=True, dropout=0.0):
        super(UNetDown, self).__init__()
        layers = [nn.Conv2d(in_size, out_size, 4, 2, 1, bias=False)]
        if normalize:
            layers.append(nn.InstanceNorm2d(out_size))
        layers.append(nn.LeakyReLU(0.2))
        if dropout:
            layers.append(nn.Dropout(dropout))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class UNetUp(nn.Module):
    def __init__(self, in_size, out_size, dropout=0.0):
        super(UNetUp, self).__init__()
        layers = [
            nn.ConvTranspose2d(in_size, out_size, 4, 2, 1, bias=False),
            nn.InstanceNorm2d(out_size),
            nn.ReLU(inplace=True),
        ]
        if dropout:
            layers.append(nn.Dropout(dropout))

        self.model = nn.Sequential(*layers)

    def forward(self, x, skip_input):
        x = self.model(x)
        x = torch.cat((x, skip_input), 1)

        return x


class GeneratorUNet(nn.Module):
    def __init__(self, in_channels=3, out_channels=3):
        super(GeneratorUNet, self).__init__()

        self.down1 = UNetDown(in_channels, 64, normalize=False)
        self.down2 = UNetDown(64, 128)
        self.down3 = UNetDown(128, 256)
        self.down4 = UNetDown(256, 512, dropout=0.5)
        self.down5 = UNetDown(512, 512, dropout=0.5)
        self.down6 = UNetDown(512, 512, dropout=0.5)
        self.down7 = UNetDown(512, 512, dropout=0.5)
        self.down8 = UNetDown(512, 512, normalize=False, dropout=0.5)

        self.up1 = UNetUp(512, 512, dropout=0.5)
        self.up2 = UNetUp(1024, 512, dropout=0.5)
        self.up3 = UNetUp(1024, 512, dropout=0.5)
        self.up4 = UNetUp(1024, 512, dropout=0.5)
        self.up5 = UNetUp(1024, 256)
        self.up6 = UNetUp(512, 128)
        self.up7 = UNetUp(256, 64)

        self.final = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(128, out_channels, 4, padding=1),
            nn.Tanh(),
        )

    def forward(self, x):
        # U-Net generator with skip connections from encoder to decoder
        d1 = self.down1(x)
        d2 = self.down2(d1)
        d3 = self.down3(d2)
        d4 = self.down4(d3)
        d5 = self.down5(d4)
        d6 = self.down6(d5)
        d7 = self.down7(d6)
        d8 = self.down8(d7)
        u1 = self.up1(d8, d7)
        u2 = self.up2(u1, d6)
        u3 = self.up3(u2, d5)
        u4 = self.up4(u3, d4)
        u5 = self.up5(u4, d3)
        u6 = self.up6(u5, d2)
        u7 = self.up7(u6, d1)

        return self.final(u7)


##############################
#        Discriminator
##############################


class DiscriminatorBlock(nn.Module):
    def __init__(self, in_filters, out_filters, normalization=True):
        super(DiscriminatorBlock, self).__init__()
        """Returns downsampling layers of each discriminator block"""
        layers = [nn.Conv2d(in_filters, out_filters, 4, stride=2, padding=1)]
        if normalization:
            layers.append(nn.BatchNorm2d(out_filters))
        layers.append(nn.LeakyReLU(0.2, inplace=True))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class Discriminator(nn.Module):
    def __init__(self, in_channels=3):
        super(Discriminator, self).__init__()

        def discriminator_block(in_filters, out_filters, normalization=True):
            """Returns downsampling layers of each discriminator block"""
            layers = [nn.Conv2d(in_filters, out_filters, 4, stride=2, padding=1)]
            if normalization:
                layers.append(nn.InstanceNorm2d(out_filters))
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            return layers

        self.model = nn.Sequential(
            *discriminator_block(in_channels * 2, 64, normalization=False),
            *discriminator_block(64, 128),
            *discriminator_block(128, 256),
            *discriminator_block(256, 512),
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(512, 1, 4, padding=1, bias=False)
        )

    def forward(self, img_A, img_B):
        # Concatenate image and condition image by channels to produce input
        img_input = torch.cat((img_A, img_B), 1)
        return self.model(img_input)


class PerceptualLossDiscriminator(nn.Module):
    def __init__(self, in_channels=3):
        super(PerceptualLossDiscriminator, self).__init__()

        self.disc_blk_1 = DiscriminatorBlock(in_channels * 2, 64, normalization=False)
        self.disc_blk_2 = DiscriminatorBlock(64, 128)
        self.disc_blk_3 = DiscriminatorBlock(128, 256)
        self.disc_blk_4 = DiscriminatorBlock(256, 512)
        self.final = nn.Sequential(
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(512, 1, 4, padding=1, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, img_A, img_B):
        # Concatenate image and condition image by channels to produce input
        img_input = torch.cat((img_A, img_B), 1)
        disc_1 = self.disc_blk_1(img_input)
        disc_2 = self.disc_blk_2(disc_1)
        disc_3 = self.disc_blk_3(disc_2)
        disc_4 = self.disc_blk_4(disc_3)
        return self.final(disc_4)


class pix2pix(nn.Module):
    def __init__(
        self,
        data,
        ch_in: int,
        ch_out: int,
        backbone=None,
        perceptual_loss: bool = False,
        chip_size=256,
    ):
        super().__init__()

        if perceptual_loss:
            self.D = PerceptualLossDiscriminator(ch_in)
        else:
            self.D = Discriminator(ch_in)
        if backbone:
            self.G = GeneratorResUnet(data, ch_out, chip_size, backbone)
        else:
            self.G = GeneratorUNet(ch_in, ch_out)

        self.perceptual_loss = perceptual_loss
        self.arcgis_results = False

    def forward(self, real_A, real_B):
        fake_B = self.G(real_A)
        if self.training:
            self.arcgis_results = False
        if self.arcgis_results:
            return torch.cat([fake_B[:, None], fake_B[:, None]], 1)
        # if not self.training:
        return [fake_B]


class AdaptiveLoss(nn.Module):
    def __init__(self, crit):
        super().__init__()
        self.crit = crit

    def forward(self, output, target: bool, **kwargs):
        targ = (
            output.new_ones(*output.size())
            if target
            else output.new_zeros(*output.size())
        )
        return self.crit(output, targ, **kwargs)


class Adaptivel1Loss(nn.Module):
    def __init__(self, crit1):
        super().__init__()
        self.crit1 = crit1

    def forward(self, output, target: bool, **kwargs):
        targ = (
            output.new_ones(*output.size())
            if target
            else output.new_zeros(*output.size())
        )
        return self.crit1(output, targ, **kwargs)


class RMSELog(nn.Module):
    def __init__(self):
        super(RMSELog, self).__init__()

    def forward(self, fake, real):
        eps = 1e-7
        real = F.relu(real)
        fake = F.relu(fake)
        loss = torch.sqrt(
            torch.mean(torch.abs(torch.log(real + eps) - torch.log(fake + eps)) ** 2)
        )
        return loss


class get_activation_from_layers:
    def __init__(self, name):
        self.name = name

    def get_activation(self):
        activation = {}

        def hook(model, input, output):
            activation[self.name] = output.detach()

        return hook, activation


class Pix2PixPerceptualLoss(nn.Module):
    def __init__(
        self,
        cgan: nn.Module,
        lambda_A: float = 100.0,
        lambda_B: float = 100.0,
        lambda_idt: float = 0.5,
        lsgan: bool = False,
    ):
        super().__init__()
        self.cgan, self.l_A, self.l_B, self.l_idt = cgan, lambda_A, lambda_B, lambda_idt
        self.crit = AdaptiveLoss(F.binary_cross_entropy_with_logits)
        self.crit1 = Adaptivel1Loss(F.mse_loss)
        self.rmse_log = RMSELog()

    def set_input(self, input):
        self.real_A, self.real_B = input

    def forward(self, output, target):
        fake_B = output[0]

        self.gen_loss = self.crit(self.cgan.D(fake_B, self.real_A), True)
        self.l1_loss = torch.mean(F.l1_loss(fake_B, self.real_B))
        self.rmse_log_loss = self.rmse_log(self.real_B, fake_B)

        activ1 = get_activation_from_layers("disc_blk_1")
        j1, k1 = activ1.get_activation()
        hook1 = self.cgan.D.disc_blk_1.register_forward_hook(j1)
        ten1 = self.cgan.D(self.real_B, self.real_B)
        k1 = k1["disc_blk_1"]
        hook1.remove()

        activ2 = get_activation_from_layers("disc_blk_1")
        j2, k2 = activ2.get_activation()
        hook2 = self.cgan.D.disc_blk_1.register_forward_hook(j2)
        ten2 = self.cgan.D(fake_B, fake_B)
        k2 = k2["disc_blk_1"]
        hook2.remove()

        activ3 = get_activation_from_layers("disc_blk_2")
        j3, k3 = activ3.get_activation()
        hook3 = self.cgan.D.disc_blk_2.register_forward_hook(j3)
        ten3 = self.cgan.D(self.real_B, self.real_B)
        k3 = k3["disc_blk_2"]
        hook3.remove()

        activ4 = get_activation_from_layers("disc_blk_2")
        j4, k4 = activ4.get_activation()
        hook4 = self.cgan.D.disc_blk_2.register_forward_hook(j4)
        ten4 = self.cgan.D(fake_B, fake_B)
        k4 = k4["disc_blk_2"]
        hook4.remove()

        activ5 = get_activation_from_layers("disc_blk_3")
        j5, k5 = activ5.get_activation()
        hook5 = self.cgan.D.disc_blk_3.register_forward_hook(j5)
        ten5 = self.cgan.D(self.real_B, self.real_B)
        k5 = k5["disc_blk_3"]
        hook5.remove()

        activ6 = get_activation_from_layers("disc_blk_3")
        j6, k6 = activ6.get_activation()
        hook6 = self.cgan.D.disc_blk_3.register_forward_hook(j6)
        ten6 = self.cgan.D(fake_B, fake_B)
        k6 = k6["disc_blk_3"]
        hook6.remove()

        percep_loss, percep_loss1, percep_loss2, percep_loss3 = 0, 0, 0, 0
        for i in range(k1.shape[0]):
            percep_loss1 += F.l1_loss(k1[i, :, :, :], k2[i, :, :, :])
            percep_loss2 += F.l1_loss(k3[i, :, :, :], k4[i, :, :, :])
            percep_loss3 += F.l1_loss(k5[i, :, :, :], k6[i, :, :, :])
            percep_loss += (
                (5 * percep_loss1) + (1.5 * percep_loss2) + (1 * percep_loss3)
            )
        self.perceptual_loss = percep_loss / 3

        return (
            self.gen_loss
            + (self.l_A * self.l1_loss)
            + self.perceptual_loss
            + self.rmse_log_loss
        )


class pix2pixLoss(nn.Module):
    def __init__(
        self,
        cgan: nn.Module,
        lambda_A: float = 100.0,
        lambda_B: float = 100,
        lambda_idt: float = 0.5,
        lsgan: bool = False,
    ):
        super().__init__()
        self.cgan, self.l_A, self.l_B, self.l_idt = cgan, lambda_A, lambda_B, lambda_idt
        self.crit = AdaptiveLoss(F.binary_cross_entropy_with_logits)
        self.crit1 = Adaptivel1Loss(F.mse_loss)

    def set_input(self, input):
        self.real_A, self.real_B = input

    def forward(self, output, target):
        fake_B = output[0]

        self.gen_loss = self.crit(self.cgan.D(fake_B, self.real_A), True)
        self.l1_loss = torch.mean(F.l1_loss(fake_B, self.real_B))

        return self.gen_loss + (self.l_A * self.l1_loss)


class pix2pixTrainer(LearnerCallback):
    _order = -20  # Need to run before the Recorder

    def _set_trainable(self, D=False):
        gen = not D
        requires_grad(self.learn.model.G, gen)
        requires_grad(self.learn.model.D, D)
        if not gen:
            self.opt_D.lr, self.opt_D.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D.wd, self.opt_D.beta = self.learn.opt.wd, self.learn.opt.beta

    def on_train_begin(self, **kwargs):
        self.G = self.learn.model.G
        self.D = self.learn.model.D
        self.crit = self.learn.loss_func.crit
        self.crit1 = self.learn.loss_func.crit1

        if not getattr(self, "opt_G", None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G))])
        else:
            self.opt_G.lr, self.opt_G.wd = self.opt.lr, self.opt.wd
            self.opt_G.mom, self.opt_G.beta = self.opt.mom, self.opt.beta

        if not getattr(self, "opt_D", None):
            self.opt_D = self.learn.opt.new([nn.Sequential(*flatten_model(self.D))])

        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.gen_smter, self.l1_smter = SmoothenValue(0.98), SmoothenValue(0.98)
        self.d_smter = SmoothenValue(0.98)
        self.recorder.add_metric_names(["gen_loss", "l1_loss", "D_loss"])

    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)

    def on_backward_begin(self, **kwargs):
        self.l1_smter.add_value(self.loss_func.l1_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())

    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G.zero_grad()
        fake_B = last_output[0].detach()
        real_A, real_B = last_input

        self._set_trainable(D=True)

        self.D.zero_grad()

        loss_D = 0.5 * (
            torch.mean(self.crit1(self.D(real_A, real_B), True))
            + torch.mean(self.crit1(self.D(fake_B, real_A), False))
        )

        self.d_smter.add_value(loss_D.detach().cpu())
        if self.learn.model.training == True:
            loss_D.backward()

        self.opt_D.step()

        self._set_trainable()

    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(
            last_metrics,
            [s.smooth for s in [self.gen_smter, self.l1_smter, self.d_smter]],
        )


class Pix2PixPerceptualTrainer(LearnerCallback):
    _order = -20  # Need to run before the Recorder

    def _set_trainable(self, D=False):
        gen = not D
        requires_grad(self.learn.model.G, gen)
        requires_grad(self.learn.model.D, D)
        if not gen:
            self.opt_D.lr, self.opt_D.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D.wd, self.opt_D.beta = self.learn.opt.wd, self.learn.opt.beta

    def on_train_begin(self, **kwargs):
        self.G = self.learn.model.G
        self.D = self.learn.model.D
        self.crit = self.learn.loss_func.crit
        self.crit1 = self.learn.loss_func.crit1

        if not getattr(self, "opt_G", None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G))])
        else:
            self.opt_G.lr, self.opt_G.wd = self.opt.lr, self.opt.wd
            self.opt_G.mom, self.opt_G.beta = self.opt.mom, self.opt.beta

        if not getattr(self, "opt_D", None):
            self.opt_D = self.learn.opt.new([nn.Sequential(*flatten_model(self.D))])

        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.gen_smter, self.l1_smter = SmoothenValue(0.98), SmoothenValue(0.98)
        self.d_smter, self.rmselog_smter = SmoothenValue(0.98), SmoothenValue(0.98)
        self.percep_smter = SmoothenValue(0.98)
        self.recorder.add_metric_names(
            ["gen_loss", "l1_loss", "D_loss", "rmse_log_loss", "perceptual_loss"]
        )

    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)

    def on_backward_begin(self, **kwargs):
        self.l1_smter.add_value(self.loss_func.l1_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())
        self.rmselog_smter.add_value(self.loss_func.rmse_log_loss.detach().cpu())
        self.percep_smter.add_value(self.loss_func.perceptual_loss.detach().cpu())

    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G.zero_grad()
        fake_B = last_output[0].detach()
        real_A, real_B = last_input

        self._set_trainable(D=True)

        self.D.zero_grad()

        activ1 = get_activation_from_layers("disc_blk_1")
        j1, k1 = activ1.get_activation()
        hook1 = self.D.disc_blk_1.register_forward_hook(j1)
        ten1 = self.D(real_B, real_B)
        k1 = k1["disc_blk_1"]
        hook1.remove()

        activ2 = get_activation_from_layers("disc_blk_1")
        j2, k2 = activ1.get_activation()
        hook2 = self.D.disc_blk_1.register_forward_hook(j2)
        ten2 = self.D(fake_B, fake_B)
        k2 = k2["disc_blk_1"]
        hook2.remove()

        activ3 = get_activation_from_layers("disc_blk_2")
        j3, k3 = activ3.get_activation()
        hook3 = self.D.disc_blk_2.register_forward_hook(j3)
        ten3 = self.D(real_B, real_B)
        k3 = k3["disc_blk_2"]
        hook3.remove()

        activ4 = get_activation_from_layers("disc_blk_2")
        j4, k4 = activ4.get_activation()
        hook4 = self.D.disc_blk_2.register_forward_hook(j4)
        ten4 = self.D(fake_B, fake_B)
        k4 = k4["disc_blk_2"]
        hook4.remove()

        activ5 = get_activation_from_layers("disc_blk_3")
        j5, k5 = activ5.get_activation()
        hook5 = self.D.disc_blk_3.register_forward_hook(j5)
        ten5 = self.D(real_B, real_B)
        k5 = k5["disc_blk_3"]
        hook5.remove()

        activ6 = get_activation_from_layers("disc_blk_3")
        j6, k6 = activ6.get_activation()
        hook6 = self.D.disc_blk_3.register_forward_hook(j6)
        ten6 = self.D(fake_B, fake_B)
        k6 = k6["disc_blk_3"]
        hook6.remove()

        percep_loss, percep_loss1, percep_loss2, percep_loss3 = 0, 0, 0, 0
        for i in range(k1.shape[0]):
            percep_loss1 += F.l1_loss(k1[i, :, :, :], k2[i, :, :, :])
            percep_loss2 += F.l1_loss(k3[i, :, :, :], k4[i, :, :, :])
            percep_loss3 += F.l1_loss(k5[i, :, :, :], k6[i, :, :, :])
            percep_loss += (
                (5 * percep_loss1) + (1.5 * percep_loss2) + (1 * percep_loss3)
            )
        percep_loss = percep_loss / 3
        loss_D = 0.5 * (
            torch.mean(self.crit(self.D(real_A, real_B), True))
            + torch.mean(self.crit(self.D(fake_B, real_A), False))
        ) + (0.1 * percep_loss)

        self.d_smter.add_value(loss_D.detach().cpu())
        if self.learn.model.training == True:
            loss_D.backward()

        self.opt_D.step()

        self._set_trainable()

    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(
            last_metrics,
            [
                s.smooth
                for s in [
                    self.gen_smter,
                    self.l1_smter,
                    self.d_smter,
                    self.rmselog_smter,
                    self.percep_smter,
                ]
            ],
        )


def compute_fid_metric(model, data):
    input_a = []
    input_b = []
    pred_b = []
    for input, target in data.valid_dl:
        input_a.append(input[0] / 2 + 0.5)
        input_b.append(input[1] / 2 + 0.5)
        pred = model.learn.pred_batch(batch=(input, target))
        if isinstance(pred, list):
            pred = pred[0]
        else:
            pred = pred[:, 0, :, :, :]
        pred_b.append(pred / 2 + 0.5)

    data_len = len(data.valid_ds)
    batch_size = data.batch_size

    m1_b, s1_b = calculate_activation_statistics(batch_size, data_len, input_b)
    m2_b, s2_b = calculate_activation_statistics(batch_size, data_len, pred_b)

    fid_value = calculate_frechet_distance(m1_b, s1_b, m2_b, s2_b)

    return fid_value


def compute_metrics(model, dl, show_progress):
    avg_psnr = 0
    avg_ssim = 0
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=False):
            prediction = model.learn.model(input[0], input[1])
            if isinstance(prediction, list):
                prediction = prediction[0]
            else:
                prediction = prediction[:, 0, :, :, :]
            avg_psnr += psnr(prediction, input[1])
            avg_ssim += ssim(prediction, input[1])
    return avg_psnr / len(dl), avg_ssim.item() / len(dl)



import json
import os
import os.path
import re
import logging

log = logging.getLogger()

import ipykernel
from notebook.notebookapp import list_running_servers
from IPython.display import HTML, display


def get_dir_of_curr_exec_notebook():
    """
    Returns the absolute path of the directory that the currently executing
    jupyter notebook resides in
    """
    try:
        kernel_id = re.search(
            "kernel-(.*).json", ipykernel.connect.get_connection_file()
        ).group(1)
        servers = list_running_servers()
        for ss in servers:
            import requests
            from requests.compat import urljoin

            response = requests.get(
                urljoin(ss["url"], "api/sessions"),
                params={"token": ss.get("token", "")},
            )
            for nn in json.loads(response.text):
                if nn["kernel"]["id"] == kernel_id:
                    relative_path = nn["notebook"]["path"]
                    nb_path = os.path.join(ss["notebook_dir"], relative_path)
                    return os.path.dirname(nb_path)
    except Exception:
        log.debug(
            "Could not discover the directory of the currently "
            "executing notebook, using os.getcwd(). To override "
            "and specify the dir manually, call `MapView.raster."
            "set_current_executing_nb_dir('/path/to/dir/'`)"
        )
        return os.getcwd()


class JupyterHorizScrollImageList:
    def __init__(self):
        self._img_el_properties = []

    def append(self, src, title):
        self._img_el_properties.append({"src": src, "title": title})

    def _ipython_display_(self):
        html = ""
        for img in self._img_el_properties:
            html += f'\n<img src="{img["src"]}" title="{img["title"]}"></img>'
        html = f'<div class="scrollmenu">{html}\n</div>'
        css = """
              div.scrollmenu {
                overflow: auto;
                white-space: nowrap;
              }
              div.scrollmenu img {
                margin-top: 1em;
                display: inline-block;
                width: 25%;
                height: auto;
                color: white;
                text-align: center;
                padding: 14px;
                text-decoration: none;
              }
              div.scrollmenu img:hover {
                background-color: #777;
              }
              """
        html = f"<style>{css}</style><html>{html}</html>"
        display(HTML(html))



"""
In the GIS, entities located in space with a set of properties can be represented as features. This module has the types
to represent features and collection of features.
"""
from __future__ import annotations
from arcgis.auth.tools import LazyLoader
from typing import Any, Optional, Union
import geomet.esri

copy = LazyLoader("copy")
json = LazyLoader("json")
_ujson = LazyLoader("ujson")
os = LazyLoader("os")
re = LazyLoader("re")
tempfile = LazyLoader("tempfile")
uuid = LazyLoader("uuid")

from datetime import datetime
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._spatial import json_to_featureclass
from arcgis._impl.common._utils import _date_handler
from arcgis.geometry import (
    BaseGeometry,
    Point,
    MultiPoint,
    Polyline,
    Polygon,
    Geometry,
    SpatialReference,
)
from arcgis.gis import Layer

try:
    arcpy = LazyLoader("arcpy", strict=True)

    HASARCPY = True
except:
    HASARCPY = False


class Feature(object):
    """Entities located in space with a set of properties can be represented as features.

    .. code-block:: python

        # Obtain a feature from a feature layer:

        # Query a Feature Layer to get a Feature Set
        >>> feature_set = feature_layer.query(where="OBJECTID=1")
        # Assign a variable to the list of features in the Feature Set
        >>> feature_list = feature_set.features
        # Get an individual feature
        >>> feature = feature_list[0]

        # Verify the object type
        >>> type(feature)
        arcgis.features.feature.Feature
        # Print the string representation of the feature
        >>> feature
        {"geometry": {"x": -8238318.738276444, "y": 4970309.724235498, "spatialReference": {"wkid": 102100, "latestWkid": 3857}},
        "attributes": {"Incident_Type": "Structural-Sidewalk Collapse", "Location": "927 Broadway", "Borough": "Manhattan",
        "Creation_Date": 1477743211000, "Closed_Date": null, "Latitude": 40.7144215406227, "Longitude": -74.0060763804198,
        "ObjectId": 1}}
    """

    _geom = None
    _json = None
    _dict = None
    _geom_type = None
    _attributes = None
    _wkid = None

    # ----------------------------------------------------------------------
    def __init__(self, geometry=None, attributes=None):
        """Constructor"""
        self._dict = {}
        if geometry is not None:
            self._dict["geometry"] = geometry
        if attributes is not None:
            self._dict["attributes"] = attributes

    # ----------------------------------------------------------------------
    def set_value(
        self,
        field_name: str,
        value: Union[dict, BaseGeometry, Point, MultiPoint, Polyline, Polygon],
    ):
        """
        Sets an attribute value for a given field name.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        field_name          Required String. The name of the field to update.
        ---------------     --------------------------------------------------------------------
        value               Required. Value to update the field with.
        ===============     ====================================================================

        :return:
            A boolean indicating whether ``field_name`` value was updated (True), or not updated (False).

        .. code-block:: python

            # Usage Example

            >>> feat_set = feature_layer.query(where="OBJECTID=1")
            >>> feat = feat_set.features[0]
            >>> feat.fields
            ['OBJECTID',
             'FID_Commun',
             'AREA',
             'PERIMETER',
             'NAME',
             'COUNTY',
             'CONAME']
            >>> feat.get_value("NAME")
            'Original Name'
            >>> feat.set_value(field_name = "NAME", value = "New Name")
            True

        .. note::
            To save edits from the above snippet, use :meth:`~arcgis.features.FeatureLayer.edit_features`
            with `feat_set` in a list as the `updates` argument.
        """

        if field_name in self.fields:
            if value is not None:
                self._dict["attributes"][field_name] = value
                self._json = json.dumps(self._dict, default=_date_handler)
            else:
                pass
        elif field_name.upper() in ["SHAPE", "SHAPE@", "GEOMETRY"]:
            if isinstance(value, BaseGeometry):
                if isinstance(value, Point):
                    self._dict["geometry"] = {
                        "x": value["x"],
                        "y": value["y"],
                    }
                elif isinstance(value, MultiPoint):
                    self._dict["geometry"] = {"points": value["points"]}
                elif isinstance(value, Polyline):
                    self._dict["geometry"] = {"paths": value["paths"]}
                elif isinstance(value, Polygon):
                    self._dict["geometry"] = {"rings": value["rings"]}
                else:
                    return False
                self._json = json.dumps(self._dict, default=_date_handler)
        else:
            return False
        return True

    # ----------------------------------------------------------------------
    def get_value(self, field_name: str):
        """
        Retrieves the value for a specified field name.

        =============       ===========================================================
        **Parameter**        **Description**
        -------------       -----------------------------------------------------------
        field_name          Required String. The name for each attribute field.

                            .. note::
                                ``feature.fields`` will return a list of all field names.
        =============       ===========================================================

        :return:
            The value for the specified attribute field of the :class:`~arcgis.features.Feature`

        """
        if field_name in self.fields:
            return self._dict["attributes"][field_name]
        elif field_name is not None and field_name.upper() in [
            "SHAPE",
            "SHAPE@",
            "GEOMETRY",
        ]:
            return self._dict["geometry"]
        return None

    # ----------------------------------------------------------------------
    @property
    def as_dict(self):
        """
        Retrieves the feature layer as a dictionary.

        :return:
            The feature as a dictionary
        """
        d = dict(self._dict)
        if "geometry" in d and d["geometry"] in [None, {}]:
            d.pop("geometry")
        if "attributes" in d and d["attributes"] in [None, {}]:
            d.pop("attributes")
        return d

    # ----------------------------------------------------------------------
    @property
    def as_row(self):
        """
        Retrieves the feature as a tuple containing two lists:

        =============     ===========================================================
        **List of:**      **Description**
        -------------     -----------------------------------------------------------
        row values        the specific attribute values and geometry for this feature
        -------------     -----------------------------------------------------------
        field names       the name for each attribute field
        =============     ===========================================================

        :return:
            A tuple of two lists: row values and field names

        """
        fields = self.fields
        row = [""] * len(fields)
        for key, val in self._attributes.items():
            row[fields.index(key)] = val
            del val
            del key
        if self.geometry is not None:
            row.append(self.geometry)
            fields.append("SHAPE@")
        return row, fields

    # ----------------------------------------------------------------------
    @property
    def geometry(self):
        """
        Get/Set the geometry of the feature, if any.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required string.
                                Values: 'Polyline' | 'Polygon' | 'Point'

                                .. note::
                                    Setting this value will override the current geometry dictionary
                                    if already present.

        ==================      ====================================================================

        :return: The feature's geometry as a dictionary.

        .. code-block:: python

            # Get the current geometry
            >>> feat_set = feature_layer.query(where="1=1")
            >>> feat_list = feat_set.features
            >>> feat = feat_list[0]
            >>> feat.geometry
            {'x': -8238318.738276444,
             'y': 4970309.724235498,
             'spatialReference': {'wkid': 102100, 'latestWkid': 3857}}
        """
        if self._geom is None:
            if "geometry" in self._dict.keys():
                self._geom = self._dict["geometry"]
            else:
                return None
        return self._geom

    @geometry.setter
    def geometry(self, value):
        """
        See main ``geometry`` property docstring
        """
        self._geom = value
        self._dict["geometry"] = value

    # ----------------------------------------------------------------------
    @property
    def attributes(self):
        """
        Get/Set the attribute values for a feature

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required dict.
        ==================      ====================================================================

        :return:
            A dictionary of feature attribute values with field names as the key

        .. code-block:: python

            #Example to set attribute values

            >>> feat_set = feature_layer.query(where="1=1")
            >>> feat_list = feat_set.features
            >>> feat = feat_list[0]
            >>> feat.attributes = {"field1 : "value", field2 : "value"}
        """

        if self._attributes is None and "attributes" in self._dict:
            self._attributes = self._dict["attributes"]
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        """
        See main ``attributes`` property docstring
        """
        self._attributes = value
        self._dict["attributes"] = value

    # ----------------------------------------------------------------------
    @property
    def fields(self):
        """
        Retrieves the attribute field names for the feature as a list of strings

        :return:
            A list of strings
        """
        if "attributes" in self._dict:
            self._attributes = self._dict["attributes"]
            return list(self._attributes.keys())
        else:
            return []

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        """
        Retrieves the geometry type of the Feature as a string.

        :return:
            The geometry type of the :class:`~arcgis.features.Feature` as a string

        """
        if self._geom_type is None:
            if self.geometry is not None:
                if hasattr(self.geometry, "type"):
                    self._geom_type = self.geometry.type
                else:
                    self._geom_type = Geometry(self.geometry)._type
            else:
                self._geom_type = "Table"
        return self._geom_type

    # ----------------------------------------------------------------------
    @classmethod
    def from_json(cls, json_str: str):
        """
        Creates a Feature object from a JSON string.

        :return:
            A :class:`~arcgis.features.Feature`

        """
        feature = _ujson.loads(json_str)
        geom = feature["geometry"] if "geometry" in feature else None
        attribs = feature["attributes"] if "attributes" in feature else None
        return cls(geom, attribs)

    # ----------------------------------------------------------------------
    @classmethod
    def from_dict(cls, feature: str, sr: Optional[dict[str, str]] = None):
        """
        Creates a Feature object from a dictionary.

        :return:
            A :class:`~arcgis.features.Feature`
        """
        geom = feature["geometry"] if "geometry" in feature else None
        if geom and sr and isinstance(geom, dict) and not "spatialReference" in geom:
            geom["spatialReference"] = sr

        attribs = feature["attributes"] if "attributes" in feature else None
        if "centroid" in feature:
            if attribs is None:
                attribs = {"centroid": feature["centroid"]}
            elif "centroid" in attribs:
                fld = "centroid_" + uuid.uuid4().hex[:2]
                attribs[fld] = feature["centroid"]
            else:
                attribs["centroid"] = feature["centroid"]
        return cls(geom, attribs)

    # ----------------------------------------------------------------------
    def __str__(self):
        """ """
        return json.dumps(self.as_dict, default=_date_handler)

    __repr__ = __str__


class FeatureSet(object):
    """
    A ``FeatureSet`` is a set of features with information about their ``fields``, ``field aliases``, ``geometry type``,
    ``spatial reference``, and more.

    ``FeatureSets`` are commonly used as input/output with several ``Geoprocessing Tools``, and can be the obtained
    through the :attr:`~arcgis.features.FeatureLayer.query` methods of feature layers.
    A FeatureSet can be combined with a layer definition to compose a FeatureCollection.

    FeatureSet contains :class:`~arcgis.features.Feature` objects, including the values for the
    fields requested by the :class:`~arcgis.gis.User` . For layers, if you request geometry
    information, the geometry of each feature is also returned in the
    FeatureSet. For tables, the FeatureSet does not include geometries.

    If a ``Spatial Reference`` is not specified at the ``FeatureSet`` level, the
    ``FeatureSet`` will assume the SpatialReference of its first feature. If
    the ``Spatial Reference`` of the first feature is also not specified, the
    spatial reference will be UnknownCoordinateSystem.
    """

    _fields = None
    _features = None
    _has_z = None
    _has_m = None
    _geometry_type = None
    _spatial_reference = None
    _object_id_field_name = None
    _global_id_field_name = None
    _display_field_name = None
    _allowed_geom_types = [
        "esriGeometryPoint",
        "esriGeometryMultipoint",
        "esriGeometryPolyline",
        "esriGeometryPolygon",
        "esriGeometryEnvelope",
    ]

    # ----------------------------------------------------------------------
    def __init__(
        self,
        features,
        fields=None,
        has_z=False,
        has_m=False,
        geometry_type=None,
        spatial_reference=None,
        display_field_name=None,
        object_id_field_name=None,
        global_id_field_name=None,
    ):
        """Constructor"""
        self._has_z = has_z
        self._has_m = has_m
        self._geometry_type = geometry_type
        self._spatial_reference = spatial_reference
        self._display_field_name = display_field_name
        self._object_id_field_name = object_id_field_name
        self._global_id_field_name = global_id_field_name

        # conversion of different inputs to a common list of feature objects
        if isinstance(features, str):
            # convert the featuresclass to a list of features
            features, fields = self._fc_to_features(dataset=features)
            if features is None:
                raise AttributeError(
                    "Feature class could not be converted to a feature set"
                )
        elif isinstance(features, list) and len(features) > 0:
            feature = features[0]
            if isinstance(feature, Feature):
                pass
                # features passed in as a list of Feature objects
            elif isinstance(feature, dict):
                # features passed in as a list of dicts
                if "attributes" in feature:
                    if "geometry" in feature:
                        features = [
                            Feature(feat["geometry"], feat["attributes"])
                            for feat in features
                        ]
                    else:
                        features = [
                            Feature(None, feat["attributes"]) for feat in features
                        ]
                elif "geometry" in feature:
                    features = [Feature(feat["geometry"], None) for feat in features]
            else:
                raise AttributeError(
                    "FeatureSet requires a list of features (as dicts or Feature objects)"
                )
        self._fields = fields
        self._features = features
        if len(features) > 0:
            feat_geom = None
            feature = features[0]
            # Check if first feature has it, else we will enter while loop
            if (
                "geometry" in feature.as_dict
            ):  # can construct features out of tables with just attributes, no geometry
                feat_geom = feature.geometry
            elif isinstance(feature, dict):
                if "geometry" in feature:
                    feat_geom = feature["geometry"]

            i = 1
            while feat_geom is None and i < len(features):
                # while feat_geom is none and we haven't gone through all features, keep going
                if (
                    "geometry" in feature.as_dict
                ):  # can construct features out of tables with just attributes, no geometry
                    feat_geom = feature.geometry
                elif isinstance(feature, dict):
                    if "geometry" in feature:
                        feat_geom = feature["geometry"]
                if feat_geom is None:
                    # get next feature and continue while loop
                    feature = features[i]
                    # get next i
                    i = i + 1

            if feat_geom is not None:
                if spatial_reference is None:
                    if "spatialReference" in feat_geom:
                        self._spatial_reference = feat_geom["spatialReference"]

                if isinstance(feat_geom, Geometry):
                    geometry = feat_geom
                else:
                    geometry = Geometry(feat_geom)

                if geometry_type is None:
                    if isinstance(geometry, Polyline):
                        self._geometry_type = "esriGeometryPolyline"
                    elif isinstance(geometry, Polygon):
                        self._geometry_type = "esriGeometryPolygon"
                    elif isinstance(geometry, Point):
                        self._geometry_type = "esriGeometryPoint"
                    elif isinstance(geometry, MultiPoint):
                        self._geometry_type = "esriGeometryMultipoint"

            # region - build fields into a dict
            if self._fields is None or len(self._fields) == 0:
                self._fields = feature.fields  # get fields from the feature if not set

            if self._fields and isinstance(
                self._fields[0], str
            ):  # as in geocoded results
                _fields = []

                for key, val in feature.attributes.items():
                    if isinstance(val, float):
                        field_type = "esriFieldTypeDouble"
                    elif isinstance(val, int):
                        field_type = "esriFieldTypeInteger"
                    else:
                        field_type = "esriFieldTypeString"

                    _fields.append(
                        {
                            "name": key,
                            "alias": key,
                            "type": field_type,
                            "sqlType": "sqlTypeOther",
                        }
                    )

                self._fields = _fields
            # endregion

            # Try to find the object ID field if not specified
            if self._object_id_field_name is None:
                # check to see if features a dict or feature object
                if isinstance(feature, Feature):
                    # Look for OBJECTID first, if it does not exist, look for FID
                    if self._fields is None or len(self._fields) == 0:
                        self._fields = (
                            feature.fields
                        )  # get fields from first feature if not set
                    for field in feature.fields:
                        if re.search("^{0}$".format("OBJECTID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                    for field in feature.fields:
                        if re.search("^{0}$".format("FID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                else:
                    for field, _ in feature.items():
                        if re.search("^{0}$".format("OBJECTID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break
                    for field, _ in feature.items():
                        if re.search("^{0}$".format("FID"), field, re.IGNORECASE):
                            self._object_id_field_name = field
                            break

            # if still none, then build the objectid field
            if not self._object_id_field_name:
                obj_field = {
                    "name": "OBJECTID",
                    "type": "esriFieldTypeOID",
                    "alias": "OBJECTID",
                    "sqlType": "sqlTypeOther",
                }

                if len(self._fields) > 0:
                    field0 = self._fields[0]
                    if isinstance(field0, str):
                        self._fields.append("OBJECTID")
                    else:
                        self._fields.append(obj_field)
                else:
                    self._fields.append(obj_field)

                self._object_id_field_name = obj_field["name"]

                counter = 0
                for feat in self._features:
                    counter += 1
                    feat.fields.append("OBJECTID")
                    if (
                        not feat.attributes
                    ):  # when features have just geometry and no attributes
                        feat.attributes = {"OBJECTID": counter}
                    else:
                        feat.attributes["OBJECTID"] = counter

            # rectify Object ID field type
            if self._object_id_field_name:
                for f in self._fields:
                    if f["name"] == self._object_id_field_name:
                        if f["type"] != "esriFieldTypeOID":
                            f["type"] = "esriFieldTypeOID"
                        break

    # ----------------------------------------------------------------------
    def __str__(self):
        """returns object as string"""
        return json.dumps(self.value, default=_date_handler)

    def __repr__(self):
        return "<{}> {} features".format(self.__class__.__name__, len(self.features))

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _fc_to_features(dataset):
        """
        Converts a dataset to a list of feature objects, if ArcPy is available

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        dataset             Required string. Path to the featureclass.
        ===============     ====================================================================


        :return: A list of :class:`~arcgis.features.Feature` objects
        """
        try:
            import arcpy

            arcpy_found = True
        except:
            arcpy_found = False
            raise AttributeError(
                "ArcPy is required to create a feature set from a feature class"
            )
        if arcpy_found:
            if not arcpy.Exists(dataset=dataset):
                raise AttributeError(
                    "Error creating FeatureSet: {0} does not exist".format(dataset)
                )

            desc = arcpy.da.Describe(dataset)
            fields = []
            date_fields = []
            for field in desc["fields"]:
                key = field.name
                if field.type == "Date":
                    field_type = "EsriFieldTypeDate"
                    date_fields.append(key)
                elif field.type == "Double":
                    field_type = "EsriFieldTypeDouble"
                elif field.type == "Integer":
                    field_type = "EsriFieldTypeInteger"
                else:
                    field_type = "EsriFieldTypeString"
                fields.append(
                    {
                        "name": key,
                        "alias": key,
                        "type": field_type,
                        "sqlType": "sqlTypeOther",
                    }
                )
            fields_names = [field["name"] for field in fields]
            non_geom_fields = copy.deepcopy(fields_names)
            features = []
            if hasattr(desc, "shapeFieldName"):
                fields.append("SHAPE@JSON")
            del desc

            with arcpy.da.SearchCursor(
                dataset, fields_names
            ) as rows:  # get field names here
                for row in rows:
                    row = list(row)
                    for date_field in date_fields:
                        date_value = row[fields_names.index(date_field)]
                        if date_value is not None:
                            date_value = int(_date_handler(date_value))

                    template = {"attributes": dict(zip(non_geom_fields, row))}
                    if "SHAPE@JSON" in fields:
                        template["geometry"] = _ujson.loads(
                            row[fields.index("SHAPE@JSON")]
                        )
                    features.append(Feature.from_dict(template))
                    del row
            fields = fields if fields else None
            return features, fields
        return None
        # ----------------------------------------------------------------------

    @property
    def value(self):
        """
        Gets the Feature Set object as a dictionary.

        :return:
            A dictionary of the :class:`~arcgis.features.FeatureSet`
        """
        val = {"features": [f.as_dict for f in self._features]}

        if self._object_id_field_name is not None:
            val["objectIdFieldName"] = self._object_id_field_name
        if self._display_field_name is not None:
            val["displayFieldName"] = self._display_field_name
        if self._global_id_field_name is not None:
            val["globalIdFieldName"] = self._global_id_field_name
        if self._spatial_reference is not None:
            val["spatialReference"] = self._spatial_reference
        if self._geometry_type is not None:
            val["geometryType"] = self._geometry_type
        if self._has_z:
            val["hasZ"] = self._has_z
        if self._has_m:
            val["hasM"] = self._has_m
        if self._fields is not None:
            val["fields"] = self._fields

        return val

    # ----------------------------------------------------------------------
    @property
    def to_json(self):
        """
        Gets the Feature Set object as a JSON string.

        :return:
            A JSON string of the :class:`~arcgis.features.FeatureSet`
        """
        return json.dumps(self.value, default=_date_handler)

    # ----------------------------------------------------------------------
    @property
    def to_geojson(self):
        """
        Gets the Feature Set object as a GeoJSON.

        :return:
            A GeoJSON object.
        """

        def esri_to_geo(esrijson):
            """converts Esri Format JSON to GeoJSON"""

            def extract(feature, esri_geom_type):
                """creates a single feature"""
                item = {}
                item["type"] = "Feature"
                geom = feature["geometry"]
                geometry = {}
                geometry["type"] = get_geom_type(esri_geom_type)
                geometry["coordinates"] = get_coordinates(geom, geometry["type"])
                # add check for MultiPolygon
                if geometry["type"] == "Polygon" and len(geometry["coordinates"]) > 1:
                    geometry["type"] = "MultiPolygon"
                item["geometry"] = geometry
                item["properties"] = feature["attributes"]

                return item

            def get_geom_type(esri_type):
                """converts esri geometry types to
                GeoJSON geometry types"""
                if esri_type == "esriGeometryPoint":
                    return "Point"
                elif esri_type == "esriGeometryMultiPoint":
                    return "MultiPoint"
                elif esri_type == "esriGeometryPolyline":
                    return "MultiLineString"
                elif esri_type == "esriGeometryPolygon":
                    return "Polygon"
                else:
                    return "Point"

            def get_coordinates(geom, geom_type):
                """
                converts the Esri Geometry Structure to
                GeoJSON structure"""
                if geom_type == "Polygon":
                    return geom["rings"]
                elif geom_type == "MultiLineString":
                    return geom["paths"]
                elif geom_type == "Point":
                    return [geom["x"], geom["y"]]
                else:
                    return []

            geojson = {}
            features = esrijson["features"]
            esri_geom_type = esrijson["geometryType"]
            geojson["type"] = "FeatureCollection"
            feats = []
            for feat in features:
                feats.append(extract(feat, esri_geom_type))
            geojson["features"] = feats
            return geojson

        return json.dumps(esri_to_geo(self.value), default=_date_handler)

    def to_dict(self):
        """
        Converts the Feature Set object to a Python dictionary.

        :return:
            A Python dictionary of the :class:`~arcgis.features.FeatureSet`
        """
        return self.value

    @property
    def df(self):
        """

        .. Warning::
            deprecated in v1.5.0 please use :attr:`~arcgis.features.FeatureSet.sdf`

        converts the FeatureSet to a Pandas dataframe. Requires pandas
        """
        return self.sdf

    # ----------------------------------------------------------------------
    @property
    def sdf(self):
        """
        Gets the Feature Set as a Spatially Enabled Pandas dataframe.

        :return:
            A Spatially Enabled `Pandas Dataframe <https://pandas.pydata.org/docs/reference/frame.html#dataframe>`_
            object
        """
        try:
            from arcgis.features.geo._io.serviceops import from_featureset

            return from_featureset(fset=self)
        except ImportError:
            raise Exception(
                "Could not find the panda installation, please install Pandas and retry"
            )
        except Exception as e:
            raise Exception(
                "An error occurred with exporting the FeatureSet with message: %s"
                % str(e)
            )

    # ----------------------------------------------------------------------
    def __iter__(self):
        """featureset iterator on features in feature set"""
        for feature in self._features:
            yield feature

    # ----------------------------------------------------------------------
    def __len__(self):
        """returns the number of features in feature set"""
        return len(self._features)

    # ----------------------------------------------------------------------
    @staticmethod
    def from_arcpy(fs: "arcpy.FeatureSet") -> "FeatureSet":
        """
        Converts an `arcpy` FeatureSet to an `arcgis` FeatureSet

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        fs                  Required arcpy.FeatureSet. The featureset objec to consume.
        ===============     ====================================================================

        :return:
           A :class:`~arcgis.features.FeatureSet` object
        """
        if hasattr(fs, "JSON") and type(fs).__name__ == "FeatureSet":
            return FeatureSet.from_dict(_ujson.loads(fs.JSON))
        else:
            raise ValueError(f"Invalid input of type {type(fs)}")

    # ----------------------------------------------------------------------
    @staticmethod
    def from_json(json_str: str):
        """
        Creates a Feature Set objects from a JSON string.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        json_str            Required json style string.
        ===============     ====================================================================

        :return:
           A :class:`~arcgis.features.FeatureSet` object
        """
        return FeatureSet.from_dict(_ujson.loads(json_str))

    @staticmethod
    def from_dataframe(df):
        """
        The ``from_dataframe`` method creates a :class:`~arcgis.features.FeatureSet` objects from a
        Pandas' DataFrame

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        df                  Required DataFrame.
        ===============     ====================================================================

        :return:
           A :class:`~arcgis.features.FeatureSet` object
        """

        def _infer_type(df, col):
            """
            internal function used to get the datatypes for the feature class if
            the dataframe's _field_reference is NULL or there is a column that does
            not have a dtype assigned to it.

            Input:
             dataframe - spatialdataframe object
            Ouput:
              field type name
            """
            import numpy as np

            nn = df[col].notnull()
            nn = list(df[nn].index)
            if len(nn) > 0:
                val = df[col][nn[0]]
                if isinstance(val, (str, pd.StringDtype)):
                    return "esriFieldTypeString"
                elif isinstance(
                    val,
                    tuple([int] + [np.int32, pd.Int32Dtype, pd.Int16Dtype]),
                ):
                    return "esriFieldTypeInteger"
                elif isinstance(val, (float, np.int64, pd.Int64Dtype)):
                    return "esriFieldTypeDouble"
                elif isinstance(val, datetime):
                    return "esriFieldTypeDate"
            return "esriFieldTypeString"

        import pandas as pd

        features = []
        index = 0
        sr = None
        try:
            date_cols = [col for col in df.columns if df[col].dtype == "datetime64[ns]"]
            cols = [
                col
                for col in df.columns
                if col != df.spatial.name and col not in date_cols
            ]
            df = df.fillna("")
        except:
            pass
        old_idx = df.index
        df.reset_index(drop=True, inplace=True)
        if isinstance(df, pd.DataFrame) and not df.spatial.name is None:
            fs = FeatureSet.from_dict(df.spatial.__feature_set__)
            df.set_index(old_idx, inplace=True)
            return fs
        elif isinstance(df, pd.DataFrame):
            geoms = []
            df_rows = df.copy()
        else:
            raise ValueError("Invalid input type")
        index = 0
        for row in df_rows.to_dict("records"):
            if len(geoms) > 0:
                features.append(
                    {
                        "geometry": _ujson.loads(json.dumps(geoms[index])),
                        "attributes": row,
                    }
                )
            else:
                features.append({"attributes": row})
            index += 1
        fs = FeatureSet.from_dict(featureset_dict={"features": features})
        fields = []
        for col in df_rows.columns:
            # if col not in df_rows.geometry.name:
            fields.append({"name": col, "type": _infer_type(df=df_rows, col=col)})
        fs._fields = fields
        if sr is not None:
            fs.spatial_reference = sr
        df.set_index(old_idx, inplace=True)
        return fs

    # ----------------------------------------------------------------------
    @staticmethod
    def from_geojson(geojson):
        """
        Creates a Feature Set objects from a GEO JSON  :class:`~arcgis.features.FeatureCollection` object

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        geojson             Required GEOJSON object
        ===============     ====================================================================

        :return:
           A :class:`~arcgis.features.FeatureSet` object

        """
        from warnings import warn

        def geo_to_esri(geojson):
            esri = {}

            # we already know the spatial reference we want
            # for geojson, at least in this simple case
            if "crs" in geojson:
                warn(
                    "crs has been deprecated and will be ignored. Please"
                    + " see: https://tools.ietf.org/html/rfc7946#section-4 for"
                    + " more information."
                )
            sr = {"wkid": 4326}

            # This will hold the geojson geometry type
            geo_type = ""
            # check for collection of features
            # and iterate as necessary
            attribute_fields = []
            if geojson["type"] == "FeatureCollection":
                features = geojson["features"]
                geo_type = features[0]["geometry"]["type"]
                attribute_fields = features[0]["properties"]

                esri_features = map(extract, features)
            else:
                attribute_fields = geojson["properties"]
                geo_type = geojson["geometry"]["type"]
                esri_features = extract(geojson)

            fields = map(extract_field, attribute_fields)
            # everything should be ready to define for the
            # esri json
            esri["geometryType"] = get_geom_type(geo_type)
            esri["spatialReference"] = sr
            esri["fields"] = list(fields)
            esri["features"] = list(esri_features)

            return esri

        def extract(feature):
            # parse out the geometry data
            geometry = get_geometry(feature)
            out_feature = {}
            out_feature["geometry"] = geometry
            out_feature["attributes"] = feature["properties"]

            return out_feature

        def extract_field(attribute):
            # now we need the fields in the properties
            a = {}
            a["alias"] = attribute
            a["name"] = attribute
            if isinstance(attribute, int):
                a["type"] = "esriFieldTypeSmallInteger"
            elif isinstance(attribute, float):
                a["type"] = "esriFieldTypeDouble"
            else:
                a["type"] = "esriFieldTypeString"
                a["length"] = 70
            return a

        def get_geom_type(geo_type):
            if geo_type == "Point":
                return "esriGeometryPoint"
            elif geo_type == "MultiPoint":
                return "esriGeometryMultiPoint"
            elif geo_type in ["LineString", "MultiLineString"]:
                return "esriGeometryPolyline"
            elif geo_type == "Polygon" or geo_type == "MultiPolygon":
                return "esriGeometryPolygon"
            else:
                return "unknown"

        def get_geometry(feature):
            # match how geometry is represented
            # based on the geojson geometry type

            geom = feature["geometry"]
            if HASARCPY:
                geom = arcpy.AsShape(geom)
                geometry = Geometry(geom)
            else:
                geometry = Geometry(geomet.esri.dumps(geom))
            return geometry

        return FeatureSet.from_dict(geo_to_esri(geojson))

    # ----------------------------------------------------------------------
    @staticmethod
    def from_dict(featureset_dict: dict[str, Any]):
        """
        Creates a Feature Set objects from a dictionary.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        featureset_dict     Required dict.
                            Keys can include:
                            'fields', 'features', 'hasZ', 'hasM', 'geometryType', 'objectIdFieldName',
                            'globalIdFieldName', 'displayFieldName', 'spatialReference'
        ===============     ====================================================================

        :return:
           A :class:`~arcgis.features.FeatureSet`
        """
        features = []
        if "fields" in featureset_dict:
            fields = featureset_dict["fields"]
        else:
            fields = []
        if "features" in featureset_dict:
            sr = featureset_dict.get("spatialReference", None)
            for feat in featureset_dict["features"]:
                features.append(Feature.from_dict(feat, sr=sr))
        return FeatureSet(
            features=features,
            fields=fields,
            has_z=featureset_dict["hasZ"] if "hasZ" in featureset_dict else False,
            has_m=featureset_dict["hasM"] if "hasM" in featureset_dict else False,
            geometry_type=featureset_dict["geometryType"]
            if "geometryType" in featureset_dict
            else None,
            object_id_field_name=featureset_dict["objectIdFieldName"]
            if "objectIdFieldName" in featureset_dict
            else None,
            global_id_field_name=featureset_dict["globalIdFieldName"]
            if "globalIdFieldName" in featureset_dict
            else None,
            display_field_name=featureset_dict["displayFieldName"]
            if "displayFieldName" in featureset_dict
            else None,
            spatial_reference=featureset_dict["spatialReference"]
            if "spatialReference" in featureset_dict
            else None,
        )

    # ----------------------------------------------------------------------
    @property
    def spatial_reference(self):
        """
        Get/Set the Feature Set's spatial reference

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required dict.
                                (e.g. {"wkid" : 4326})
        ==================      ====================================================================

        :return:
            A :class:`~arcgis.geometry.SpatialReference`

        """
        return self._spatial_reference

    # ----------------------------------------------------------------------
    @spatial_reference.setter
    def spatial_reference(self, value: Optional[dict[str, int]]):
        """
        See main ``spatial_reference`` property docstring
        """
        if isinstance(value, SpatialReference):
            self._spatial_reference = value
        elif isinstance(value, int):
            self._spatial_reference = SpatialReference(wkid=value)
        elif isinstance(value, str) and str(value).isdigit():
            self._spatial_reference = SpatialReference(wkid=int(value))
        else:
            self._spatial_reference = SpatialReference(value)

    # ----------------------------------------------------------------------
    @property
    def has_z(self):
        """
         Get/Set the Z-property of the Feature Set object

         ==================      ====================================================================
         **Parameter**            **Description**
         ------------------      --------------------------------------------------------------------
         value                   Required bool.
                                 Values: True | False
         ==================      ====================================================================

        :return:
             The Z-value of the :class:`~arcgis.features.FeatureSet` object
        """
        return self._has_z

    # ----------------------------------------------------------------------
    @has_z.setter
    def has_z(self, value: Optional[bool]):
        """
        See main ``has_z`` property docstring
        """
        if isinstance(value, bool):
            self._has_z = value

    # ----------------------------------------------------------------------
    @property
    def has_m(self):
        """
        Get/Set the M-property of the Feature Set object.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required bool.
                                Values: True | False
        ==================      ====================================================================

        :return:
            The M-value of the :class:`~arcgis.features.FeatureSet` object
        """
        return self._has_m

    # ----------------------------------------------------------------------
    @has_m.setter
    def has_m(self, value: Optional[bool]):
        """
        See main ``has_m`` property docstring
        """
        if isinstance(value, bool):
            self._has_m = value

    # ----------------------------------------------------------------------
    @property
    def geometry_type(self):
        """
        Get/Set the ``Type`` of the Feature Set object.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required string.
                                Values: 'Polygon' | 'Polyline' | 'Point'
        ==================      ====================================================================

        :return:
            A string representing the geometry type of the :class:`~arcgis.features.FeatureSet` object
        """
        return self._geometry_type

    # ----------------------------------------------------------------------
    @geometry_type.setter
    def geometry_type(self, value: Optional[str]):
        """
        See main ``geometry_type`` property docstring
        """
        if value in self._allowed_geom_types:
            self._geometry_type = value

    # ----------------------------------------------------------------------
    @property
    def object_id_field_name(self):
        """
        Get/Set the object id field of the Feature Set object

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required string.
        ==================      ====================================================================

        :return:
            A string representing the object id field name
        """
        return self._object_id_field_name

    # ----------------------------------------------------------------------
    @object_id_field_name.setter
    def object_id_field_name(self, value: Optional[str]):
        """
        See main ``object_id_field_name`` property docstring
        """
        self._object_id_field_name = value

    # ----------------------------------------------------------------------
    @property
    def global_id_field_name(self):
        """
        Get/Set the ``global ID`` field  for the Feature Set object.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required string.
        ==================      ====================================================================


        :return:
            A string
        """
        return self._global_id_field_name

    # ----------------------------------------------------------------------
    @global_id_field_name.setter
    def global_id_field_name(self, value: Optional[str]):
        """
        See main ``global_id_field_name`` property docstring
        """
        self._global_id_field_name = value

    # ----------------------------------------------------------------------
    @property
    def display_field_name(self):
        """
        Get/Set the ``display`` field for the Feature Set object.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required string.
        ==================      ====================================================================

        :return:
            A String
        """
        return self._display_field_name

    # ----------------------------------------------------------------------
    @display_field_name.setter
    def display_field_name(self, value: Optional[str]):
        """
        See main ``display_field_name`` property docstring
        """
        self._display_field_name = value

    # ----------------------------------------------------------------------
    def save(
        self,
        save_location: str,
        out_name: str,
        encoding: Optional[str] = None,
    ):
        """
        The ``save`` method saves a Feature Set object to a
        :class:`~arcgis.features.Feature` class on disk.

        =================    ====================================================================
        **Parameter**         **Description**
        -----------------    --------------------------------------------------------------------
        save_location        Required string. Path to export the Feature Set to.
        -----------------    --------------------------------------------------------------------
        out_name             Required string. Name of the saved table.
        -----------------    --------------------------------------------------------------------
        encoding             Optional string. character encoding is used to represent a
                             repertoire of characters by some kind of encoding system. The
                             default is None.
        =================    ====================================================================

        :return:
            A string

        .. code-block:: python

            # Obtain a feature from a feature layer:

            >>> feat_set = feature_layer.save(save_location = "C:\ArcGISProjects\"
            >>>                               out_name = "Power_Plant_Data")
            "C:\ArcGISProjects\Power_Plant_Data"

        """
        _, file_extension = os.path.splitext(out_name)
        if file_extension.lower() not in [".csv", ".json"] and HASARCPY == False:
            raise ImportError("ArcPy is required to export a feature class.")
        import sys

        if sys.version_info[0] == 2:
            access = "wb+"
            kwargs = {}
            if encoding is not None:
                kwargs["encoding"] = encoding
        else:
            access = "wt+"
            kwargs = {"newline": ""}
            if encoding is not None:
                kwargs["encoding"] = encoding

        if file_extension == ".csv":
            res = os.path.join(save_location, out_name)
            with open(res, access, **kwargs) as csv_file:
                import csv

                csv_writer = csv.writer(csv_file)
                fields = []
                # write the headers to the csv
                for field in self.fields:
                    fields.append(field["name"])
                csv_writer.writerow(fields)

                new_row = []
                # Loop through the results and save each to a row
                for feature in self:
                    new_row = []
                    for field in self.fields:
                        new_row.append(feature.get_value(field["name"]))
                    csv_writer.writerow(new_row)
                csv_file.close()
            del csv_file
        elif file_extension == ".json":
            res = os.path.join(save_location, out_name)
            with open(res, access, **kwargs) as writer:
                json.dump(
                    self.value,
                    writer,
                    sort_keys=True,
                    indent=4,
                    ensure_ascii=False,
                )
            del writer
        else:
            temp_dir = tempfile.gettempdir()
            temp_file = os.path.join(temp_dir, "%s.json" % uuid.uuid4().hex)
            with open(temp_file, access, **kwargs) as writer:
                json.dump(self.value, writer, default=_date_handler)
            del writer
            res = json_to_featureclass(
                json_file=temp_file,
                out_fc=os.path.join(save_location, out_name),
            )
            os.remove(temp_file)
        return res

    # ----------------------------------------------------------------------
    @property
    def features(self):
        """
        Gets the :class:`~arcgis.features.Feature` objects in the FeatureSet object.

        :return:
            A list of :class:`~arcgis.features.Feature` objects
        """
        return self._features

    # ----------------------------------------------------------------------
    @property
    def fields(self):
        """
        Get/Set the fields in the FeatureSet

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        value                   Required dict.
        ==================      ====================================================================


        :return:
            A dictionary
        """
        # Build object id field if not found - webmaps need this
        if not self._object_id_field_name:
            obj_field = {
                "name": "OBJECTID",
                "type": "esriFieldTypeOID",
                "alias": "OBJECTID",
                "sqlType": "sqlTypeOther",
            }

            if len(self._fields) > 0:
                field0 = self._fields[0]
                if isinstance(field0, str):
                    self._fields.append("OBJECTID")
                else:
                    self._fields.append(obj_field)
            else:
                self._fields.append(obj_field)

            self._object_id_field_name = obj_field["name"]

            counter = 0
            for feat in self.features:
                counter += 1
                feat.fields.append("OBJECTID")
                feat.attributes["OBJECTID"] = counter

        return self._fields

    # ----------------------------------------------------------------------
    @fields.setter
    def fields(self, fields: Optional[dict]):
        """
        See main ``fields`` property docstring
        """
        self._fields = fields


class FeatureCollection(Layer):
    """
    ``FeatureCollection`` is an object with a layer definition and a :class:`~arcgis.features.FeatureSet`.

    It is an in-memory collection of :class:`~arcgis.features.Feature` objects with rendering information.

    .. note::
        Feature Collections can be stored as :class:`~arcgis.gis.Item` objects in the GIS, added as layers to a map or
        scene, passed as inputs to feature analysis tools, and returned as results from feature analysis tools
        if an output name for a feature layer is not specified when calling the tool.
    """

    # noinspection PyMissingConstructor
    def __init__(self, dictdata):
        self._hydrated = True
        self.properties = PropertyMap(dictdata)
        self.layer = self.properties

    @property
    def _lyr_json(self):
        return dict(self.properties)

    @property
    def _lyr_dict(self):
        return dict(self.properties)

    def __str__(self):
        return "<%s>" % type(self).__name__

    def __repr__(self):
        return "<%s>" % type(self).__name__

    def query(self):
        """
        Retrieves the data in this feature collection as a :class:`~arcgis.features.FeatureSet`.
        Ex: FeatureCollection.query()

        .. Warning::
            Filtering by ``where clause`` is not supported for feature collections.

        :return:
            A :class:`~arcgis.features.FeatureSet` object

        """
        if "layers" in self.properties:
            if "fields" in self.properties["layers"][0]["layerDefinition"]:
                self.properties["layers"][0]["featureSet"]["fields"] = self.properties[
                    "layers"
                ][0]["layerDefinition"]["fields"]

            return FeatureSet.from_dict(
                self.properties["layers"][0]["featureSet"],
            )
        else:
            if "fields" in self.properties["layerDefinition"]:
                self.properties["featureSet"]["fields"] = self.properties[
                    "layerDefinition"
                ]["fields"]

            return FeatureSet.from_dict(self.properties["featureSet"])

    @staticmethod
    def from_featureset(
        fset: FeatureSet,
        symbol: Optional[dict[str, Any]] = None,
        name: Optional[str] = None,
    ):
        """
        Creates a :class:`~arcgis.features.FeatureCollection` object from a :class:`~arcgis.features.FeatureSet` object.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        fset                   Required :class:`~arcgis.features.FeatureSet` object.
        ------------------     --------------------------------------------------------------------
        symbol                 Optional dict. Specify your symbol as a dictionary. Symbols for points
                               can be picked from the
                               `Esri Symbol Page <http://esri.github.io/arcgis-python-api/tools/symbol.html>`_

                               If not specified, a default symbol will be created.
        ------------------     --------------------------------------------------------------------
        name                   Optional String. The name of the feature collection. This is used
                               when feature collections are being persisted on a WebMap. If None is
                               provided, then a random name is generated. (New at 1.6.1)
        ==================     ====================================================================

        :return:
            A :class:`~arcgis.features.FeatureCollection` object.

        .. code-block:: python

            # Usage Example

            >>> feat_set = feature_layer.query(where="OBJECTID=1")
            >>> feat_collect = FeatureCollection.from_featureset(feat_set)
            >>> type(feat_collect)
            "acrgis.features.FeatureCollection"


        """
        if not isinstance(fset, FeatureSet):
            raise ValueError

        fset_dict = fset.to_dict()

        # region compose layer definition

        fc_layer_definition = {
            "geometryType": fset_dict["geometryType"],
            "fields": fset_dict["fields"],
            "spatialReference": fset_dict["spatialReference"],
            "objectIdField": fset.object_id_field_name,
            "type": "Feature Layer",
        }

        if not "name" in fc_layer_definition:
            if name:
                fc_layer_definition["name"] = name.replace(" ", "_")
            else:
                fc_layer_definition["name"] = "a" + uuid.uuid4().hex[:5]

        if not "id" in fc_layer_definition:
            fc_layer_definition["id"] = 0

        if not symbol:
            if fc_layer_definition["geometryType"] == "esriGeometryPolyline":
                symbol = {
                    "color": [0, 0, 0, 255],
                    "width": 1.33,
                    "type": "esriSLS",
                    "style": "esriSLSSolid",
                }

            elif fc_layer_definition["geometryType"] in [
                "esriGeometryPolygon",
                "esriGeometryEnvelope",
            ]:
                symbol = {
                    "color": [0, 0, 0, 64],
                    "outline": {
                        "color": [0, 0, 0, 255],
                        "width": 1.33,
                        "type": "esriSLS",
                        "style": "esriSLSSolid",
                    },
                    "type": "esriSFS",
                    "style": "esriSFSSolid",
                }

            elif fc_layer_definition["geometryType"] in [
                "esriGeometryPoint",
                "esriGeometryMultipoint",
            ]:
                symbol = {
                    "type": "esriSMS",
                    "color": [226, 29, 145, 158],
                    "angle": 0,
                    "xoffset": 0,
                    "yoffset": 0,
                    "size": 12,
                    "style": "esriSMSCircle",
                    "outline": {
                        "type": "esriSLS",
                        "color": [0, 0, 0, 255],
                        "width": 0.75,
                        "style": "esriSLSSolid",
                    },
                }

        fc_layer_definition["drawingInfo"] = {
            "renderer": {"type": "simple", "symbol": symbol}
        }
        # endregion
        # compose the feature collection dict

        layers_dict = {
            "featureSet": {
                "geometryType": fset_dict["geometryType"],
                "features": fset_dict["features"],
            },
            "layerDefinition": fc_layer_definition,
        }

        fc_dict = {"layers": [layers_dict]}

        # create a FC and return
        return FeatureCollection(fc_dict)



from requests.auth import (
    _basic_auth_str,
    HTTPBasicAuth,
    HTTPDigestAuth,
    HTTPProxyAuth,
)
from ._schain import SupportMultiAuth
from ..tools._lazy import LazyLoader
from ..tools import parse_url, assemble_url

requests = LazyLoader("requests")


###########################################################################
class ProxyAuth(HTTPProxyAuth, SupportMultiAuth):  # pragma: no cover
    """Extends ProxyAuth for Chaining"""

    def __init__(self, username: str, password: str):  # pragma: no cover
        HTTPProxyAuth.__init__(username, password)


###########################################################################
class DigestAuth(HTTPDigestAuth, SupportMultiAuth):  # pragma: no cover
    """Implements HTTPDigestAuth for chaining"""

    def __init__(self, username: str, password: str):
        HTTPDigestAuth.__init__(self, username, password)

    @property
    def token(self) -> str:
        """
        returns the authentication token
        """
        return None


###########################################################################
class EsriBasicAuth(HTTPBasicAuth, SupportMultiAuth):
    """Describes a basic requests authentication."""

    _server_log = None
    auth = None

    def __init__(
        self,
        username: str,
        password: str,
        referer: str = "http",
        verify_cert: bool = True,
        **kwargs,
    ):
        self.username = username
        self.password = password
        self._server_log = dict()
        self._tokens = dict()
        self.referer = referer or ""
        self.verify_cert = verify_cert
        self._session = requests.Session()
        self._session.verify = verify_cert
        self._session.headers.update({"referer": referer})
        self._session.auth = (self.username, self.password)
        self._proxies = kwargs.pop("proxies", None)

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __eq__(self, other):
        return all(
            [
                self.username == getattr(other, "username", None),
                self.password == getattr(other, "password", None),
            ]
        )

    def __ne__(self, other):
        return not self == other

        # ----------------------------------------------------------------------

    @property
    def token(self) -> str:
        """
        Gets the token.  This is always `None` for `EsriBasicAuth`

        :returns: String
        """
        return None

    def generate_portal_server_token(self, r, **kwargs):
        """generates a server token using Portal token"""
        parsed = parse_url(r.url)
        server_url: str = assemble_url(parsed)

        if (
            r.text.lower().find("invalid token") > -1
            or r.text.lower().find("token required") > -1
            or r.text.lower().find("unauthorized") > -1
        ) or server_url in self._server_log:
            expiration = 16000

            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": self.referer or "http",
                "f": "json",
            }
            if expiration:
                postdata["expiration"] = expiration
            if server_url in self._server_log:
                token_url = self._server_log[server_url]
            else:
                info = self._session.get(
                    server_url + "/rest/info?f=json",
                    auth=self._session.auth,
                    verify=self.verify_cert,
                    proxies=self._proxies,
                    timeout=5,
                ).json()
                token_url = info["authInfo"]["tokenServicesUrl"]
                self._server_log[server_url] = token_url
            if server_url in self._tokens:
                token_str = self._tokens[server_url]
            else:
                token = self._session.post(
                    token_url,
                    data=postdata,
                    auth=self._session.auth,
                    verify=self.verify_cert,
                    proxies=self._proxies,
                )
                token_str = token.json().get("token", None)
                if token_str is None:
                    return r
                self._tokens[server_url] = token_str
            # Recreate the request with the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self.referer or "http"
            r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r.history.append(r)
            return _r
        return r

    def __call__(self, r):
        if self.auth:
            self.auth.__call__(r)
        r.headers["Authorization"] = _basic_auth_str(self.username, self.password)
        r.register_hook("response", self.generate_portal_server_token)
        return r
